/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#ifndef MULTIGRID_CLIENT_INCLUDED
#define MULTIGRID_CLIENT_INCLUDED
#include "LaplacianMatrix/SocketedStreamingSolver.h"
#include "MultigridThread.h"
#include "SocketData.h"

template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
class SocketedMultigridClient
{
	in_addr _myAddress;
	int _processCount;
	ProcessingBlockData* _threadBlockData;
	ProcessConnectionInfo* _pConnectInfo;
	ClientData _clientData;
	int _start,_end,_width,_height,_cEnd,_cHeight,_dCount;
	bool _inCore;
	GlobalData _globalData;
public:
	SocketedMultigridClient(void);
	~SocketedMultigridClient(void);
	bool SetUp( char* serverAddress , char* prefix , int port , int index , int w , int h , bool inCore=false );
	void Run( StreamingGrid* pixels , StreamingGrid* labels , StreamingGrid* out , MultiStreamIOServer* ioServer , bool showProgress );
	int width	(void)	const;	// the width of the image
	int cWidth	(void)	const;	// the clipped width of the image
	int size	(void)	const;	// the width of the band
	int cSize	(void)	const;	// the clipped width of the band
	int height	(void)	const;	// the height of the image/band
	int cHeight	(void)	const;	// the clipped height of the image/band
	int start	(void)	const;
	int end		(void)	const;
	int cEnd	(void)	const;

	int quality	(void)	const;
	int iters	(void)	const;
	int vCycles	(void)	const;
	bool sharpen(void)	const;
	bool hdr	(void)	const;
	int         tileWidth    ( void ) const;
	int         tileHeight   ( void ) const;
	const char* tileExtension( void ) const;
	int processCount( void ) const;
};

template< int Channels , class StorageType , class PixelType , class LabelType >
class SocketedSuperMultigridClient
{
	char* _address;
	char* _prefix;
	int _port;
	int _clientNum;
	bool _inCore;
	int _blockSize;
	GlobalData _globalData;
	ClientData _clientData;
	template< class SyncType > void _Run( StreamingGrid* pixels , StreamingGrid* labels , StreamingGrid* out );
	MultiStreamIOServer* sharedIOServer;
public:
	bool SetUp( char* address , char* prefix , int port , int index , int w , int h , int subClients , bool inCore , MultiStreamIOServer* ioServer );
	void Run( StreamingGrid* pixels , StreamingGrid* labels , StreamingGrid* out );
	int size	(void)	const;	// the width of the band
	int cSize	(void)	const;	// the clipped width of the band
	int height	(void)	const;	// the height of the image/band
	int cHeight	(void)	const;	// the clipped height of the image/band
	int quality	(void)	const;
	bool sharpen(void)	const;
	bool hdr	(void)	const;
	int         tileWidth    ( void ) const;
	int         tileHeight   ( void ) const;
	const char* tileExtension( void ) const;
};
#include "MultigridProcess.inl"
#endif // MULTIGRID_CLIENT_INCLUDED