/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <winsock2.h>
#include <ws2tcpip.h>
#include "Util/MultiStreamIO.h"
#include "Util/Socket.h"
#include "Util/Half/half.h"

#define SOCKET_CONNECT_WAIT 500

/////////////////////////////
// SocketedMultigridClient //
/////////////////////////////
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::SocketedMultigridClient(void)
{
	_threadBlockData = NULL;
	_pConnectInfo = NULL;
	_processCount = 0;
}
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::~SocketedMultigridClient(void)
{
	if( _threadBlockData )
	{
		for( int i=0 ; i<_processCount ; i++ )
		{
			if( _threadBlockData[i].outHighR ) delete _threadBlockData[i].outHighR , _threadBlockData[i].outHighR = NULL;
			if( _threadBlockData[i].inHighB )  delete _threadBlockData[i].inHighB  , _threadBlockData[i].inHighB  = NULL;
			if( _threadBlockData[i].outLowR )  delete _threadBlockData[i].outLowR  , _threadBlockData[i].outLowR  = NULL;
			if( _threadBlockData[i].inLowX )   delete _threadBlockData[i].inLowX   , _threadBlockData[i].inLowX   = NULL;
		}
		delete[] _threadBlockData , _threadBlockData = NULL;
	}
	if( _pConnectInfo ) delete[] _pConnectInfo , _pConnectInfo = NULL;
	_processCount = 0;
}
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
bool SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::sharpen( void ) const { return _globalData.sharpen; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
bool SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::hdr( void ) const { return _globalData.hdr; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::tileWidth( void ) const { return _globalData.tileWidth; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::tileHeight( void ) const { return _globalData.tileHeight; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
const char* SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::tileExtension( void ) const { return _globalData.tileExt; }

template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::quality( void ) const { return _globalData.quality; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::cSize( void ) const { return _clientData.cEnd - _clientData.start; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::size( void ) const { return _clientData.end - _clientData.start; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::cHeight( void ) const { return _globalData.cHeight; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::height( void ) const { return _globalData.height; }
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
int SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::processCount( void ) const { return _processCount; }

template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
bool SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::SetUp( char* serverAddress , char* prefix , int port , int index , int w , int h , bool inCore )
{
	int pid = GetCurrentProcessId( );
	SOCKET serverSocket = GetConnectSocket( serverAddress , port , SOCKET_CONNECT_WAIT , false );

	bool haveAddress = false;
	if( prefix )
	{
		char hostAddress[512];
		GetHostAddress( hostAddress , prefix );
		if( strstr( hostAddress , prefix ) ) inet_aton( hostAddress , &_myAddress ) , haveAddress = true;
		else fprintf( stderr , "Warning: Failed to find address starting with: %s\n" , prefix );
	}
	if( !haveAddress )
	{
		sockaddr_in local;
		int foo = sizeof( local );
		getsockname( serverSocket , (sockaddr*) &local , &foo );
		memcpy( &_myAddress , &local.sin_addr , sizeof( local.sin_addr ) );
	}
	// Send the band information
	_clientData.index	= index;
	_clientData.width	= w;
	_clientData.height	= h;
	_inCore =inCore;
	SendOnSocket( serverSocket , GetPointer( _clientData ) , sizeof(_clientData) , "Failed to send client data to server" );

	// Get the global information
	ReceiveOnSocket( serverSocket , GetPointer( _globalData ) , sizeof(_globalData) , "Failed to receive global data from server ");

	// Get the local information
	ReceiveOnSocket( serverSocket , GetPointer( _clientData ) , sizeof(_clientData) , "Failed to receive client data from server" );

	// Get back the total number of threads this client is responsible for and the information about these threads
	ReceiveOnSocket( serverSocket , GetPointer( _processCount ) , sizeof(_processCount) , "Failed to receive process count from server" );
	Pointer( ProcessData ) processData = AllocArray< ProcessData >( _processCount , 1 , "SocketedMultigridClient::SetUp (processData)" );
	ReceiveOnSocket( serverSocket , processData , sizeof(ProcessData) * _processCount , "Failed to receive process data from server" );

	// At the point the client should not have to communicate with the server any more
	CloseSocket( serverSocket );

	// Start constructing the threads that will run the different processes
	_threadBlockData = new ProcessingBlockData[ _processCount ];
	IPData *addressX	= new IPData[_processCount];
	IPData *addressB	= new IPData[_processCount];
	IPData *addressLeft	= new IPData[_processCount];
	for( int i=0 ; i<_processCount ; i++ )
	{
		_threadBlockData[i].pData = processData[i];
		memcpy( &addressX[i].address    , &_myAddress , sizeof( _myAddress ) );
		memcpy( &addressB[i].address    , &_myAddress , sizeof( _myAddress ) );
		memcpy( &addressLeft[i].address , &_myAddress , sizeof( _myAddress ) );
	}
	FreeArray( processData );


	_pConnectInfo = new ProcessConnectionInfo[ _processCount ];
	int listenerCount = 0;
	HANDLE* listenerHandles = new HANDLE[ _processCount*3 ];
	SOCKET *leftSockets = new SOCKET[ _processCount ];


	// Establish the connection between each process and the server
	for( int i=0 ; i<_processCount ; i++ )
	{
		_threadBlockData[i].serverSocket = GetConnectSocket( serverAddress , port , SOCKET_CONNECT_WAIT , false );
		SendOnSocket( _threadBlockData[i].serverSocket , GetPointer( _threadBlockData[i].pData.index  ) , sizeof( _threadBlockData[i].pData.index ) , "Failed to send thread index to server" );
		SendOnSocket( _threadBlockData[i].serverSocket , GetPointer( _threadBlockData[i].pData.offset ) , sizeof( _threadBlockData[i].pData.offset ) , "Failed to send thread offset to server" );
		if( _threadBlockData[i].pData.index )	// If not at the leaf
		{
			_pConnectInfo[i].init( _threadBlockData[i].pData.children );
			_pConnectInfo[i].listenChildX = GetListenSocket( addressX[i].port );
			if ( _pConnectInfo[i].listenChildX == INVALID_SOCKET ) return false;
			SendOnSocket ( _threadBlockData[i].serverSocket , GetPointer( addressX[i] ) , sizeof(addressX[i]) , "Failed to send X address to server" );
			listenerHandles[listenerCount] = SpawnListenerSocket( _pConnectInfo[i].listenChildX , _pConnectInfo[i].childX , _threadBlockData[i].pData.children );
			if( !listenerHandles[listenerCount] ) return false;
			listenerCount++;

			_pConnectInfo[i].listenChildB = GetListenSocket( addressB[i].port );
			if ( _pConnectInfo[i].listenChildB == INVALID_SOCKET ) return false;
			SendOnSocket ( _threadBlockData[i].serverSocket , GetPointer( addressB[i] ) , sizeof(addressB[i]) , "Failed to send B address to server" );
			listenerHandles[listenerCount] = SpawnListenerSocket( _pConnectInfo[i].listenChildB , _pConnectInfo[i].childB , _threadBlockData[i].pData.children );
			if( !listenerHandles[listenerCount] ) return false;
			listenerCount++;
		}
		else _pConnectInfo[i].listenChildX = _pConnectInfo[i].listenChildB = INVALID_SOCKET;

		if( _globalData.spherical || _threadBlockData[i].pData.offset ) // If spherical or not left
		{
			_pConnectInfo[i].listenLeft = GetListenSocket( addressLeft[i].port );
			if ( _pConnectInfo[i].listenLeft == INVALID_SOCKET ) return false;
			SendOnSocket ( _threadBlockData[i].serverSocket , GetPointer( addressLeft[i] ), sizeof(addressLeft[i]) , "Failed to send left address to server" );
			listenerHandles[listenerCount] = SpawnListenerSocket( _pConnectInfo[i].listenLeft , &leftSockets[i] , 1 );
			if( !listenerHandles[listenerCount] ) return false;
			listenerCount++;
		}
		else leftSockets[i] = INVALID_SOCKET;
	}
	
	// Get connection information from the server and make the connection
	for( int i=0 ; i<_processCount ; i++ )
	{
		IPData neighborData;

		ReceiveOnSocket( _threadBlockData[i].serverSocket , GetPointer( neighborData ) , sizeof(neighborData) , "Failed to get parent X address from server" );
		_pConnectInfo[i].parentX = GetConnectSocket( neighborData.address , neighborData.port , SOCKET_CONNECT_WAIT , false );
		ReceiveOnSocket( _threadBlockData[i].serverSocket , GetPointer( neighborData ) , sizeof(neighborData) , "Failed to receive parent B address from server" );
		_pConnectInfo[i].parentB = GetConnectSocket( neighborData.address , neighborData.port , SOCKET_CONNECT_WAIT , false );

		if( _globalData.spherical || _threadBlockData[i].pData.offset < _threadBlockData[i].pData.maxOffset )
		{
			ReceiveOnSocket( _threadBlockData[i].serverSocket , GetPointer( neighborData ) , sizeof(IPData) , "Failed to get right address from server" );
			_pConnectInfo[i].right = new Socket( neighborData.address , neighborData.port , SOCKET_CONNECT_WAIT , false );
		}
		else _pConnectInfo[i].right = NULL;
	}
	if( listenerCount )
		MyWaitForMultipleObjects( listenerCount , listenerHandles , 10000 , "SocketedMultigridClient<Channels,StorageType,PixelType,LabelType>::SetUp" );
	for( int i=0 ; i<_processCount ; i++ ) if( leftSockets[i]!=INVALID_SOCKET ) _pConnectInfo[i].left = new Socket( leftSockets[i] );


	if( _globalData.spherical )
	{
		{
			int i=0;
			int depths = _threadBlockData[i].pData.endDepth - _threadBlockData[i].pData.startDepth + 1;
			Pointer( int ) ports = AllocArray< int >( depths+1 , 1 , "SocketedMultigridClient::SetUp (ports)" );
			_threadBlockData[0].syncSockets = new SOCKET[depths+1];	
			ReceiveOnSocket ( _threadBlockData[i].serverSocket , ports , sizeof(int) , "Failed to get port from server" );
			ReceiveOnSocket ( _threadBlockData[i].serverSocket , ports+1 , sizeof(int)*depths , "Failed to get ports from server" );

			for( int j=0 ; j<depths+1 ; j++ )
			{
				_threadBlockData[i].syncSockets[j] = GetConnectSocket( serverAddress , ports[j] , SOCKET_CONNECT_WAIT , false );
				if( _threadBlockData[i].syncSockets[j] == INVALID_SOCKET )	return false;
			}
			FreeArray( ports );
		}
		for( int i=1 ; i<_processCount ; i++ )
		{
			int depths = _threadBlockData[i].pData.endDepth - _threadBlockData[i].pData.startDepth + 1;
			Pointer( int ) ports = AllocArray< int >( depths , 1 , "SocketedMultigridClient::SetUp (ports)" );
			_threadBlockData[i].syncSockets = new SOCKET[depths];
			ReceiveOnSocket ( _threadBlockData[i].serverSocket , ports , sizeof(int)*depths , "Failed to get ports from server" );
			for( int j=0 ; j<depths ; j++ )
			{
				_threadBlockData[i].syncSockets[j] = GetConnectSocket( serverAddress , ports[j] , SOCKET_CONNECT_WAIT , false );
				if( _threadBlockData[i].syncSockets[j] == INVALID_SOCKET )	return false;
			}
			FreeArray( ports );
		}
	}
	else for( int i=0 ; i<_processCount ; i++ ) _threadBlockData[i].syncSockets = NULL;
	delete[] listenerHandles;
	delete[] addressX;
	delete[] addressB;
	delete[] addressLeft;
	delete[] leftSockets;
	return true;
}
template< int Channels, class StorageType , class SyncType , class PixelType , class LabelType >
void SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType >::Run( StreamingGrid* pixels , StreamingGrid* labels , StreamingGrid* out , MultiStreamIOServer* multiStreamIOServer , bool showProgress )
{
	// Initialize the processes
	DotProductStencil dotMajor , d2DotMajor , dotMinor , d2DotMinor;
	MultigridThread< Channels , StorageType , SyncType , PixelType , LabelType > thread;
	thread.pixels = pixels;
	thread.labels = labels;
	_threadBlockData[0].leftStream  = _pConnectInfo[0].left;
	_threadBlockData[0].rightStream = _pConnectInfo[0].right;
	_threadBlockData[0].outHighX = out;
	_threadBlockData[0].outHighR = NULL;


	for( int i=1 ; i<_processCount ; i++ )
	{
		_threadBlockData[i].leftStream  = _pConnectInfo[i].left;
		_threadBlockData[i].rightStream = _pConnectInfo[i].right;
	}
	int pid = GetCurrentProcessId( );
	for( int i=0 ; i<_processCount ; i++ )
		if( _threadBlockData[i].rightStream )
		{
			_threadBlockData[i].rightStream->write( ( Pointer( byte ) )GetPointer( _myAddress ) , sizeof(_myAddress) );
			_threadBlockData[i].rightStream->write( ( Pointer( byte ) )GetPointer( pid ) , sizeof(pid) );
		}
	for( int i=0 ; i<_processCount ; i++ )
		if( _threadBlockData[i].leftStream )
		{
			struct in_addr leftAddress;
			int leftPID;
			_threadBlockData[i].leftStream->read( ( Pointer( byte ) )GetPointer( leftAddress ) , sizeof(leftAddress) );
			_threadBlockData[i].leftStream->read( ( Pointer( byte ) )GetPointer( leftPID ) , sizeof(leftPID) );

			if( _myAddress.s_addr==leftAddress.s_addr && pid==leftPID )
			{
				SharedMemoryBuffer::StreamPair sPair;
				SharedMemoryBuffer::StreamPair::CreateSharedBufferPair( sPair );
				_threadBlockData[i].leftStream->write( ( Pointer( byte ) )GetPointer( sPair.second ) , sizeof( sPair.second ) );
				delete _threadBlockData[i].leftStream;
				_pConnectInfo[i].left = sPair.first;
				_threadBlockData[i].leftStream = sPair.first;
			}
			else
			{
				SharedMemoryBuffer::SecondStream* sStream = NULL;
				_threadBlockData[i].leftStream->write( ( Pointer( byte ) )GetPointer( sStream ) , sizeof( sStream ) );
			}
		}
	for( int i=0 ; i<_processCount ; i++ )
		if( _threadBlockData[i].rightStream )
		{
			SharedMemoryBuffer::SecondStream* sStream;
			_threadBlockData[i].rightStream->read( ( Pointer( byte ) )GetPointer( sStream ) , sizeof(sStream ) );
			if( sStream )
			{
				delete _threadBlockData[i].rightStream;
				_pConnectInfo[i].right = sStream;
				_threadBlockData[i].rightStream = sStream;
			}
		}

	for( int i=0 ; i<_processCount ; i++ )
	{
		int width  = _threadBlockData[i].pData.stop-_threadBlockData[i].pData.start;
		int height = _threadBlockData[i].pData.height;
		width  >>= _threadBlockData[i].pData.endDepth-_threadBlockData[i].pData.startDepth;
		height >>= _threadBlockData[i].pData.endDepth-_threadBlockData[i].pData.startDepth;

		SendOnSocket( _pConnectInfo[i].parentB , GetPointer( _threadBlockData[i].pData.offset ) , sizeof(_threadBlockData[i].pData.offset) , "Failed to send offset to parent B" );
		int width2 = width/2;
		SendOnSocket( _pConnectInfo[i].parentB , GetPointer( width2 ) , sizeof(int) , "failed to send width to parent B" );
		SendOnSocket( _pConnectInfo[i].parentX , GetPointer( _threadBlockData[i].pData.offset ) , sizeof(_threadBlockData[i].pData.offset) , "Failed to send offset to parent X" );
		SendOnSocket( _pConnectInfo[i].parentX , GetPointer( width ) , sizeof(int) , "Failed to send width to parent X" );

		_threadBlockData[i].outLowR = new SocketBackedGrid( _pConnectInfo[i].parentB , width * Channels * sizeof( SyncType ) /2 , height/2 , false , true );
		_threadBlockData[i].inLowX  = new SocketBackedGrid( _pConnectInfo[i].parentX , width * Channels * sizeof( SyncType )    , height   , false , true );
	}
	for( int i=1 ; i<_processCount ; i++ )
	{
		_threadBlockData[i].inHighB  = GetMultiSocketBackedGrid< Channels , SyncType > ( _pConnectInfo[i].childB , _threadBlockData[i].pData.children , _threadBlockData[i].pData.height    , false , true );
		_threadBlockData[i].outHighR = GetMultiSocketBackedGrid< Channels , SyncType > ( _pConnectInfo[i].childX , _threadBlockData[i].pData.children , _threadBlockData[i].pData.height<<1 , false , true );
	}

	thread.Initialize( multiStreamIOServer , _threadBlockData , _processCount , _globalData , (_globalData.progress && showProgress) , _inCore , dotMajor , d2DotMajor , dotMinor , d2DotMinor );
	thread.RunThread( &thread );
	for( int i=0; i<_processCount ; i++ )
	{
		if( _pConnectInfo[i].left ) delete _pConnectInfo[i].left;
		if( _pConnectInfo[i].right ) delete _pConnectInfo[i].right;
		CloseSocket( _threadBlockData[i].serverSocket );
		CloseSocket( _pConnectInfo[i].parentX );
		CloseSocket( _pConnectInfo[i].parentB );
		if( _pConnectInfo[i].childX ) for( int j=0 ; j<_pConnectInfo[i].children ; j++ ) CloseSocket( _pConnectInfo[i].childX[j] );
		if( _pConnectInfo[i].childB ) for( int j=0 ; j<_pConnectInfo[i].children ; j++ ) CloseSocket( _pConnectInfo[i].childB[j] );


		if( _globalData.spherical )
			for( int i=0 ; i<_processCount ; i++ )
			{
				int depths = _threadBlockData[i].pData.endDepth - _threadBlockData[i].pData.startDepth + 1;
				if( !i ) for( int j=0 ; j<depths+1 ; j++ ) CloseSocket( _threadBlockData[i].syncSockets[j] );
				else	 for( int j=0 ; j<depths   ; j++ ) CloseSocket( _threadBlockData[i].syncSockets[j] );
			}
	}
}
//////////////////////////////////
// SocketedSuperMultigridClient //
//////////////////////////////////
template< int Channels , class StorageType , class PixelType , class LabelType >
bool SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::SetUp( char* address , char* prefix , int port , int index , int w , int h , int subClients , bool inCore , MultiStreamIOServer* server )
{
	_address = address;
	_prefix = prefix;
	_port = port;
	_clientNum = subClients;
	_inCore = inCore;

	sharedIOServer = server;
	printf( "          Image Size: [%d x %d]\n" , w , h ) , fflush( stdout );
	printf( "             Address: %s:%d\n" , address , port ) , fflush( stdout );
	printf( "Connecting to server:" ) , fflush( stdout );
	SOCKET serverSocket  = GetConnectSocket( address , port , SOCKET_CONNECT_WAIT , true );

	// Send the band information
	_clientData.index	= index;
	_clientData.width	= w;
	_clientData.height	= h;
	_clientData.subClients = subClients;
	SendOnSocket( serverSocket , GetPointer( _clientData ) , sizeof(_clientData) , "Failed to send client data to server" );

	// Get the local information
	ReceiveOnSocket( serverSocket , GetPointer( _globalData ) , sizeof(_globalData) , "Failed to get global data from server" );
	ReceiveOnSocket( serverSocket , GetPointer( _clientData ) , sizeof(_clientData) , "Failed to get client data from server" );
	ReceiveOnSocket( serverSocket , GetPointer( _blockSize  ) , sizeof(_blockSize)  , "Failed to get block size from server" );
	CloseSocket( serverSocket );

	if( (size()/_blockSize)*_blockSize != size() )
		fprintfId( stderr , "Band width is not a multiple of the block size: %d , %d\n", size() , _blockSize );
	return true;
}
template< int Channels, class StorageType , class PixelType , class LabelType >
void SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::Run( StreamingGrid* pixels , StreamingGrid* labels , StreamingGrid* out )
{
	if( _inCore )
	{
		MemoryBackedGrid *_pixels = NULL , *_labels = NULL , *_out = NULL;
		_pixels = new MemoryBackedGrid( pixels->rowSize() , pixels->rows() , pixels->SeparateColors() );
		if( labels ) _labels = new MemoryBackedGrid( labels->rowSize() , labels->rows() , labels->SeparateColors() );
		_out = new MemoryBackedGrid( out->rowSize() , out->rows() , out->SeparateColors() );
		for( int i=0 ; i<pixels->rows() ; i++ ) memcpy( (*_pixels)[i] , (*pixels)[i] , pixels->rowSize() ) , pixels->advance();
		if( labels ) for( int i=0 ; i<labels->rows() ; i++ ) memcpy( (*_labels)[i] , (*labels)[i] , labels->rowSize() ) , labels->advance();
		_Run< float >( _pixels , _labels , _out );
		for( int i=0 ; i<out->rows() ; i++ ) memcpy( (*out)[i] , (*_out)[i] , out->rowSize() ) , out->advance();
		delete _pixels;
		if( _labels ) delete _labels;
		delete _out;
	}
	else
		_Run< float >( pixels , labels , out );
}
template< int Channels, class StorageType , class PixelType , class LabelType >
template< class SyncType > 
void SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::_Run( StreamingGrid* pixels , StreamingGrid* labels , StreamingGrid* out )
{
	class SocketedMultigridClientData
	{
	public:
		SocketedMultigridClientData( void ) { pixels = labels = out = NULL; }
		~SocketedMultigridClientData( void )
		{
			if( pixels ) delete pixels;
			if( labels ) delete labels;
			if( out ) delete out;
			pixels = labels = NULL;
		}
		SharedMemoryGrid *pixels , *labels , *out;
		SocketedMultigridClient< Channels , StorageType , SyncType , PixelType , LabelType > client;
		MultiStreamIOServer* ioServer;
		bool showProgress;

		char *address , *prefix;
		int port , index , width , cWidth , height , cHeight;
		bool inCore;

		static DWORD WINAPI ProcessThread( LPVOID lpParam )
		{
			SocketedMultigridClientData* data = (SocketedMultigridClientData*)lpParam;
			data->client.Run( data->pixels , data->labels , data->out , data->ioServer , data->showProgress );
			return 0;
		}
		HANDLE spawnProcess( void )
		{
			HANDLE processThread = NULL;
			DWORD processThreadID;
			processThread = CreateThread
				( 
				NULL,				// default security attributes
				0,					// use default stack size  
				ProcessThread,		// thread function 
				this,				// argument to thread function 
				0,					// use default creation flags 
				&processThreadID	// returns the thread identifier
				);
			if( !processThread )
			{
				fprintf( stderr , "CreateThread failed: " );
				PrintError();
			}
			return processThread;
		}
		static DWORD WINAPI SetUpThread( LPVOID lpParam )
		{
			SocketedMultigridClientData* data = (SocketedMultigridClientData*)lpParam;
			data->client.SetUp( data->address , data->prefix , data->port , data->index , data->width , data->height , data->inCore );
			return 0;
		}
		HANDLE spawnSetUp( void )
		{
			HANDLE setUpThread = NULL;
			DWORD setUpThreadID;
			setUpThread = CreateThread
				( 
				NULL,			// default security attributes
				0,				// use default stack size  
				SetUpThread,	// thread function 
				this,			// argument to thread function 
				0,				// use default creation flags 
				&setUpThreadID	// returns the thread identifier
				);
			if( !setUpThread )
			{
				fprintf( stderr , "CreateThread failed: " );
				PrintError();
			}
			return setUpThread;
		}
	};
	HANDLE* clientHandles = new HANDLE[ _clientNum ];
	SocketedMultigridClientData* clientData = new SocketedMultigridClientData[_clientNum];

	int blockNum = size() / _blockSize;

	int widthSum = 0;
	for( int i=0 ; i<_clientNum ; i++ )
	{
		clientData[i].address = _address;
		clientData[i].prefix = _prefix;
		clientData[i].port = _port;
		clientData[i].index = _clientData.index+i;
		clientData[i].height = height();
		clientData[i].cHeight = cHeight();
		clientData[i].inCore = _inCore;
		clientData[i].cWidth = clientData[i].width = ( ( ( blockNum * (i+1) ) / _clientNum ) - ( ( blockNum * i ) / _clientNum ) ) * _blockSize;
		widthSum += clientData[i].width;
		if( widthSum>cSize() )
		{
			clientData[i].cWidth = clientData[i].width - ( widthSum - cSize() );
			if( clientData[i].cWidth <= 0 )
			{
				fprintf( stderr , "Warning effective client width set to zero!\n" );
				clientData[i].cWidth = 0;
			}

		}
		clientHandles[i] = clientData[i].spawnSetUp();
	}

	MyWaitForMultipleObjects( _clientNum , clientHandles , 10000 , "SocketedSuperMultigridClient<Channels,StorageType,PixelType,LabelType>::_Run" );
	int pCount = 0;
	for( int i=0 ; i<_clientNum ; i++ ) pCount += clientData[i].client.processCount();
	for( int i=0 ; i<_clientNum ; i++ ) clientData[i].ioServer = sharedIOServer;
	for( int i=0 ; i<_clientNum ; i++ ) clientData[i].showProgress = false;
	clientData[0].showProgress = true;
	SharedMemoryGrid *pixelReader = NULL;
	SharedMemoryGrid *labelReader = NULL;
	SharedMemoryGrid *outWriter   = NULL;
	Pointer( PixelType ) pixelRow = NullPointer< PixelType >( );
	Pointer( LabelType ) labelRow = NullPointer< LabelType >( );
	Pointer( SyncType )    outRow = NullPointer<  SyncType >( );
	HANDLE *pixelLoadEvent = NULL , *pixelUnloadEvent = NULL;
	HANDLE *labelLoadEvent = NULL , *labelUnloadEvent = NULL;
	HANDLE   *outLoadEvent = NULL ,   *outUnloadEvent = NULL;
	int offset;

	pixelLoadEvent   = new HANDLE[ _clientNum ];
	pixelUnloadEvent = new HANDLE[ _clientNum ];
	pixelRow = AllocArray< PixelType >( Channels * cSize() , 1 , "SocketedSuperMultigridClient::pixelRow" );
	offset = 0;
	for( int i=0 ; i<_clientNum ; i++ )
	{
		pixelLoadEvent[i]   = CreateEvent( NULL , false , false , NULL );
		pixelUnloadEvent[i] = CreateEvent( NULL , false , true  , NULL );
		if( !pixelLoadEvent[i] || !pixelUnloadEvent[i] ) fprintfId( stderr , "Failed to create pixel event\n" ) , exit(0);
		clientData[i].pixels = new SharedMemoryGrid( pixelLoadEvent+i , pixelUnloadEvent+i , 1 , ( Pointer( byte ) )(pixelRow+offset) , clientData[i].cWidth * Channels * sizeof( PixelType ) , cHeight() );
		offset += clientData[i].cWidth * Channels;
	}
	pixelReader = new SharedMemoryGrid( pixelUnloadEvent , pixelLoadEvent , _clientNum , ( Pointer( byte ) )pixelRow , cSize() * Channels * sizeof( PixelType ) , cHeight() );
	pixelReader->reset( false , 1 );
	if( labels )
	{
		labelLoadEvent   = new HANDLE[ _clientNum ];
		labelUnloadEvent = new HANDLE[ _clientNum ];
		labelRow = AllocArray< LabelType >( Channels * cSize() , 1 , "SocketedSuperMultigridClient::labelRow" );
		offset = 0;
		for( int i=0 ; i<_clientNum ; i++ )
		{
			labelLoadEvent[i]   = CreateEvent( NULL , false , false , NULL );
			labelUnloadEvent[i] = CreateEvent( NULL , false , true  , NULL );
			if( !labelLoadEvent[i] || !labelUnloadEvent[i] ) fprintfId( stderr , "Failed to create label event\n" ) , exit(0);
			clientData[i].labels = new SharedMemoryGrid( labelLoadEvent+i , labelUnloadEvent+i , 1 , ( Pointer( byte ) )(labelRow+offset) , clientData[i].cWidth*Channels*sizeof(LabelType) , cHeight() );
			offset += clientData[i].cWidth*Channels;
		}
		labelReader = new SharedMemoryGrid( labelUnloadEvent , labelLoadEvent , _clientNum , ( Pointer( byte ) )labelRow , cSize()*Channels*sizeof(LabelType) , cHeight() );
		labelReader->reset( false , 1 );
	}

	outLoadEvent   = new HANDLE[ _clientNum ];
	outUnloadEvent = new HANDLE[ _clientNum ];
	outRow = AllocArray< SyncType >( Channels * size() , 1 , "SocketedSuperMultigridClient::outRow" );
	offset = 0;
	for( int i=0 ; i<_clientNum ; i++ )
	{
		outLoadEvent[i]   = CreateEvent( NULL , false , false , NULL );
		outUnloadEvent[i] = CreateEvent( NULL , false , true  , NULL );
		if( !outLoadEvent[i] || !outUnloadEvent[i] ) fprintfId( stderr , "Failed to create out event\n" ) , exit(0);
		clientData[i].out = new SharedMemoryGrid( outUnloadEvent+i , outLoadEvent+i , 1 , ( Pointer( byte ) )(outRow+offset) , clientData[i].width*Channels*sizeof( SyncType ) , height() );
		offset += clientData[i].width*Channels;
	}
	outWriter = new SharedMemoryGrid( outLoadEvent , outUnloadEvent , _clientNum , ( Pointer( byte ) )outRow , size()*Channels*sizeof( SyncType ) , height() );
	outWriter->reset( true , 1 );
	for( int i=0 ; i<_clientNum ; i++ ) clientHandles[i] = clientData[i].spawnProcess();
	for( int j=0 ; j<cHeight() ; j++ )
	{
		Pointer( PixelType ) pOut = ( Pointer( PixelType ) )( (*pixelReader)[j] );
		{
			Pointer( PixelType ) pIn = ( Pointer( PixelType ) )(*pixels)[j];
			if( pixels->SeparateColors() && !pixelReader->SeparateColors() )
				for( int c=0 ; c<Channels ; c++ )
					for( int x=0 ; x<cSize() ; x++ )
						pOut[x*Channels+c] = pIn[cSize()*c+x];
			else memcpy( pOut , pIn , sizeof( PixelType ) * Channels * cSize() );
			pixels->advance();
		}
		pixelReader->advance();
		if( labels )
		{
			Pointer( LabelType ) lOut = ( Pointer( LabelType ) )(*labelReader)[j];
			{
				Pointer( LabelType ) lIn = ( Pointer( LabelType ) )(*labels)[j];
				memcpy( lOut , lIn , sizeof( LabelType ) * Channels * cSize() );
				labels->advance();
			}
			labelReader->advance();
		}
	}
	for( int j=0 ; j<height() ; j++ )
	{
		Pointer( SyncType ) oIn = ( Pointer( SyncType ) )(*outWriter)[j];
		{
			Pointer( SyncType ) oOut = ( Pointer( SyncType ) )(*out)[j];
			int offset = 0;
			for( int i=0 ; i<_clientNum ; i++ )
			{
				int w = clientData[i].width;
				for( int x=0 ; x<w ; x++ ) for( int c=0 ; c<Channels ; c++ ) oOut[size()*c + offset +x ] = oIn[Channels*offset + w*c + x];
				offset += w;
			}
			out->advance();
		}
		outWriter->advance();
	}
	MyWaitForMultipleObjects( _clientNum , clientHandles , 10000 , "SocketedSuperMultigridClient<Channels,StorageType,PixelType,LabelType>::_Run" );
	delete[] clientHandles;
	// Delete client data before deleting all the arrays it depends on
	delete[] clientData;
	FreeArray( pixelRow );
	FreeArray( labelRow );
	FreeArray( outRow );
	if( pixelLoadEvent )
	{
		for( int i=0 ; i<_clientNum ; i++ ) CloseHandle( pixelLoadEvent[i] );
		delete[] pixelLoadEvent;
		pixelLoadEvent = NULL;
	}
	if( pixelUnloadEvent )
	{
		for( int i=0 ; i<_clientNum ; i++ ) CloseHandle( pixelUnloadEvent[i] );
		delete[] pixelUnloadEvent;
		pixelUnloadEvent = NULL;
	}
	if( labelLoadEvent )
	{
		for( int i=0 ; i<_clientNum ; i++ ) CloseHandle( labelLoadEvent[i] );
		delete[] labelLoadEvent;
		labelLoadEvent = NULL;
	}
	if( labelUnloadEvent )
	{
		for( int i=0 ; i<_clientNum ; i++ ) CloseHandle( labelUnloadEvent[i] );
		delete[] labelUnloadEvent;
		labelUnloadEvent = NULL;
	}
	if( outLoadEvent )
	{
		for( int i=0 ; i<_clientNum ; i++ ) CloseHandle( outLoadEvent[i] );
		delete[] outLoadEvent;
		outLoadEvent = NULL;
	}
	if( outUnloadEvent )
	{
		for( int i=0 ; i<_clientNum ; i++ ) CloseHandle( outUnloadEvent[i] );
		delete[] outUnloadEvent;
		outUnloadEvent = NULL;
	}
	if( pixelReader ) delete pixelReader;
	if( labelReader ) delete labelReader;
	if( outWriter )   delete outWriter;
}

template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::quality( void ) const { return _globalData.quality; }
template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::cSize( void ) const { return _clientData.cEnd - _clientData.start; }
template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::size( void ) const { return _clientData.end - _clientData.start; }
template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::cHeight( void ) const { return _globalData.cHeight; }
template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::height( void ) const { return _globalData.height; }
template< int Channels, class StorageType , class PixelType , class LabelType >
bool SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::sharpen( void ) const { return _globalData.sharpen; }
template< int Channels, class StorageType , class PixelType , class LabelType >
bool SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::hdr( void ) const { return _globalData.hdr; }
template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::tileWidth( void ) const { return _globalData.tileWidth; }
template< int Channels, class StorageType , class PixelType , class LabelType >
int SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::tileHeight( void ) const { return _globalData.tileHeight; }
template< int Channels, class StorageType , class PixelType , class LabelType >
const char* SocketedSuperMultigridClient< Channels , StorageType , PixelType , LabelType >::tileExtension( void ) const { return _globalData.tileExt; }
