/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#ifndef MULTIGRID_SERVER_INCLUDED
#define MULTIGRID_SERVER_INCLUDED
#include "LaplacianMatrix/SocketedStreamingSolver.h"
#include "MultigridThread.h"
#include "SocketData.h"

class ClientSocket
{
public:
	struct in_addr address;
	int port;
	SOCKET client;
	ClientData clientData;

	ClientSocket(void)
	{
		client = INVALID_SOCKET;
		port = 0;
	}
	~ClientSocket(void)
	{
		CloseSocket( client );
	}
	static int Sort(const void* v1,const void* v2)
	{
		ClientSocket *s1=(ClientSocket*)v1;
		ClientSocket *s2=(ClientSocket*)v2;
		return s1->clientData.index-s2->clientData.index;
	}
};
template< int Channels >
class SphericalSynchronizer
{
	int _width,_cCount;
	ClientSocket* _clientSockets;
	int _rPasses,_pPasses,_vCycles;
public:

	SphericalSynchronizer(void);
	~SphericalSynchronizer(void);
	void init(int width,const int* widths,int cCount,int rPasses,int pPasses,int vCycles );
	int clients(void)		const;
	int port(int cIndex)	const;
	template<class DataType>	void Run(void);

	template<class DataType> static DWORD WINAPI RunThread(LPVOID lpParam);
};

class ProcessSocket
{
public:
	IPData childAddressX , childAddressB , leftAddress;
	SOCKET sock;
	int index , offset;

	ProcessSocket(void)
	{
		sock = INVALID_SOCKET;
	}
	~ProcessSocket(void)
	{
		CloseSocket(sock);
	}
	static int Sort(const void* v1,const void* v2)
	{
		ProcessSocket *s1=(ProcessSocket*)v1;
		ProcessSocket *s2=(ProcessSocket*)v2;
		return s1->index-s2->index;
	}
};

template< int Channels , class SyncType >
class SocketedMultigridServer
{
	bool _noCG;
	int _oocLevels , _icLevels , _inCoreRes , _minMGRes;
	GlobalData _globalData;
	ProcessPartitionHierarchy< ProcessSocket > processPartition;
	ProcessConnectionInfo _inCoreConnectInfo;
	double _average[Channels];
	std::vector< SphericalSynchronizer< Channels >* > _oocSynchronizers;
	SphericalSynchronizer< Channels > *_icSynchronizers , imageSynchronizer;
	SOCKET* _icSyncSockets;
	HANDLE* _synchronizerHandles;


	void SolveInCore(
		DotProductStencil& dotMajor,DotProductStencil& d2DotMajor,DotProductStencil& dotMinor,DotProductStencil& d2DotMinor,
		Vector<float>& in,Vector<float>& out,double average[Channels]);

public:
	SocketedMultigridServer(void);
	~SocketedMultigridServer(void);
	bool SetUp( char* address , int port , int clientCount , int iters , int inCoreRes , int minMGRes , int vCycles , int minBandSize ,
		int tileWidth , int tileHeight , const char* tileExt , int quality , bool hdr , int lanes , bool verbose , bool spherical , bool sharpen , double iWeight , double gScale , bool blackOut , bool showProgress ,
		int* clipDimensions );
	bool SetUp( char* address , SOCKET listenSocket , int clientCount , int iters , int inCoreRes , int minMGRes , int vCycles , int minBandSize ,
		int tileWidth , int tileHeight , const char* tileExt , int quality , bool hdr , int lanes , bool verbose , bool spherical , bool sharpen , double iWeight , double gScale , bool blackOut , bool showProgress ,
		int* clipDimensions );
	void Run(void);
};

template< int Channels >
class SocketedSuperMultigridServer
{
	int _port , _clientCount , _iters , _inCoreRes ,  _minMGRes , _vCycles , _quality , _lanes , _minBandSize , _tileWidth , _tileHeight;
	bool _verbose , _spherical , _sharpen , _blackOut , _showProgress , _hdr;
	double _iWeight , _gScale;
	char _address[512] , _tileExt[16];
public:
	bool SetUp( char* prefix , int port , int clientCount , int iters , int inCoreRes , int minMGRes , int vCycles , int minBandSize ,
		int tileWidth , int tileHeight , const char* tileExt , int quality , bool hdr ,  int lanes , bool verbose , bool spherical , bool sharpen , double iWeight , double gScale , bool blackOut , bool showProgress );
	bool Run( void );
	template< class SyncType > bool _Run( void );
};
#include "MultigridServer.inl"
#endif // MULTIGRID_SERVER_INCLUDED