/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#ifndef MULTIGRID_PROCESS_INCLUDED
#define MULTIGRID_PROCESS_INCLUDED

#include "LaplacianMatrix/SocketedStreamingSolver.h"
#include "LaplacianMatrix/SocketedMultigrid/SocketData.h"

#define STREAMING_GRID_BUFFER_MULTIPLIER 2

template< class PData >
class ProcessPartition
{
public:
	class MyData
	{
	public:
		int width;
		std::vector< int > children;
		PData data;
	};
	int startDepth , endDepth;
	std::vector< MyData > processData;
	void setBounds( int idx , int& start , int& stop );
	int size( void ) const;
	void resize( int );
	MyData& operator[] ( const int& idx );
	const MyData& operator[] ( const int& idx ) const;
	static bool IsValidBandSize( int width , int height , int iters , int minSize );
};

template< class PData >
class ProcessPartitionHierarchy
{
public:
	std::vector< ProcessPartition< PData > > levels;
	int size( void ) const;
	void resize( int );
	ProcessPartition< PData >& operator[] ( const int& idx );
	const ProcessPartition< PData >& operator[] ( const int& idx ) const;
	template< class TData >
	bool Initialize( const ProcessPartition< TData >& initialPartition , int height , int iters , int depths , bool repeat , int minSize );
	int leaves( const int& depth , const int& offset ) const;
};

class ProcessingBlockData
{
public:
	ProcessData pData;
	SOCKET serverSocket , *syncSockets;
	DataStream *leftStream , *rightStream;
	StreamingGrid *outHighX , *outHighR , *inHighB , *outLowR , *inLowX;
	ProcessingBlockData( void )
	{
		outHighX = outHighR = inHighB = outLowR = inLowX = NULL;
		syncSockets = NULL;
		serverSocket = INVALID_SOCKET;
		leftStream = rightStream = NULL;
	}
};

template< int Channels , class StorageType , class SyncType , class PixelType , class LabelType >
class MultigridThread
{
	Pointer( SolverInfo< Channels > ) _solverInfo;

	SocketedStreamingDivergence< Channels , PixelType , LabelType , StorageType , SyncType >* _sDivergence;
	SocketedStreamingLaplacian < Channels , PixelType ,             StorageType , SyncType >* _sLaplacian;
	SocketedMultiGridRestrictionNode< Channels >* _sRestriction;
	SocketedMultiGridStreamingSolver< Channels , StorageType , SyncType >** _solvers;
	int _vCycles;
	bool _verbose , _outOfCore , _spherical;
	ProcessingBlockData* _blockData;
	int _blockCount;
	void _init(	ProcessingBlockData* blockData , int blockCount , const class GlobalData& globalData , bool showProgress , bool inCore , bool pixels , bool labels );
	MultiStreamIOServer* multiStreamIOServer;
public:
	StreamingGrid *pixels , *labels;
	MultigridThread( void );
	~MultigridThread( void );

	void Initialize(
		MultiStreamIOServer* ioServer ,
		ProcessingBlockData* blockData , int blockCount ,
		const GlobalData& globalData , bool showProgress , bool inCore ,
		DotProductStencil& dotMajor , DotProductStencil& d2DotMajor , DotProductStencil& dotMinor , DotProductStencil& d2DotMinor );

	static DWORD WINAPI RunThread(LPVOID lpParam);
};

#include "MultigridThread.inl"
#endif // MULTIGRID_PROCESS_INCLUDED