/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <Util/Time.h>
#include <Util/cmdLineParser.h>
#include "LaplacianMatrix/SocketedMultigrid/MultigridServer.h"
#include "LaplacianMatrix/SocketedMultigrid/SocketData.h"

#define DEFAULT_BUFLEN 1024

cmdLineInt Count( "count" ) , Port( "port" , 0 ) , Iters( "iters" , 5 ) , InCoreRes( "inCoreRes" , 1024 ) , MinMGRes( "minMGRes" , 64 ) , VCycles( "vCycles" , 1 ) , Quality( "quality" , 100 ) , Lanes( "lanes" , 1 );
cmdLineInt MinBandSize( "minBandSize" , 0 ) , TileWidth( "tileWidth" , 8192 ) , TileHeight( "tileHeight" , 8192 );
cmdLineReadable Verbose( "verbose" ) , Spherical( "spherical" ) , Progress( "progress" ) , NoBlackOut( "noBlackOut" ) , HDR( "hdr" );
cmdLineString Prefix( "prefix" ) , TileExtension( "tileExt" );
cmdLineFloat IWeight( "iWeight" , 0 ) , GScale( "gScale" , 1 );
cmdLineReadable* params[]=
{
	&Port , &Count , &Iters , &InCoreRes , &MinMGRes , &VCycles , &Verbose , &Spherical , &Progress , &Quality , &Lanes , &MinBandSize , &Prefix , &NoBlackOut ,
	&IWeight , &GScale , &HDR , &TileWidth , &TileHeight , &TileExtension ,
};

void ShowUsage(char* ex)
{
	printf("Usage %s:\n",ex) , fflush( stdout );
	printf("\t--%s <client count>\n",Count.name) , fflush( stdout );
	printf( "\t[--%s <listen port>=%d]\n" , Port.name , Port.value) , fflush( stdout );
	printf( "\t[--%s <sharpening interpolation weight>=%f]\n" , IWeight.name , IWeight.value ) , fflush( stdout );
	printf( "\t[--%s <sharpening gradient weight>=%f]\n" , GScale.name , GScale.value ) , fflush( stdout );
	printf( "\t[--%s <Gauss-Seidel iterations>=%d]\n" , Iters.name , Iters.value ) , fflush( stdout );
	printf( "\t[--%s <minimum multigrid resolution>=%d]\n" , MinMGRes.name , MinMGRes.value ) , fflush( stdout );
	printf( "\t[--%s <in core resolution>=%d]\n" , InCoreRes.name , InCoreRes.value ) , fflush( stdout );
	printf( "\t[--%s <image quality>=%d]\n" , Quality.name , Quality.value ) , fflush( stdout );

	printf( "\t[--%s <default output tile width>=%d]\n" , TileWidth.name , TileWidth.value ) , fflush( stdout );
	printf( "\t[--%s <default output tile height>=%d]\n" , TileHeight.name , TileHeight.value ) , fflush( stdout );
	printf( "\t[--%s <default output file extension>]\n" , TileExtension.name ) , fflush( stdout );

	printf( "\t[--%s]\n" , HDR.name ) , fflush( stdout );
	printf( "\t[--%s]\n" , Verbose.name ) , fflush( stdout );
	printf( "\t[--%s]\n" , Progress.name ) , fflush( stdout );
	printf( "\t[--%s]\n" , NoBlackOut.name ) , fflush( stdout );
#if 0
	printf( "\t[--%s <preferred address prefix>]\n" , Prefix.name ) , fflush( stdout );
	printf( "\t[--%s <v-cycles>=%d]\n",VCycles.name,VCycles.value) , fflush( stdout );
	printf( "\t[--%s <solver lanes>=%d]\n",Lanes.name,Lanes.value) , fflush( stdout );
	printf( "\t[--%s]\n" , Spherical.name) , fflush( stdout );
	printf( "\t[--%s]\n" , MinBandSize.name ) , fflush( stdout );
#endif
}
int main(int argc, char* argv[])
{
	MyWinSock::Load();
	int paramNum=sizeof(params)/sizeof(cmdLineReadable*);
	if( !cmdLineParse( argc-1 , &argv[1] , paramNum , params , true ) ) return EXIT_FAILURE;

	if(!Count.set)
	{
		ShowUsage(argv[0]);
		return EXIT_FAILURE;
	}
	SocketedSuperMultigridServer<3> server;
	if( !TileExtension.set )
	{
		if( !server.SetUp( Prefix.value , Port.value , Count.value , Iters.value , InCoreRes.value , MinMGRes.value , VCycles.value , MinBandSize.value , TileWidth.value , TileHeight.value , NULL , Quality.value , HDR.set , Lanes.value , Verbose.set , Spherical.set , IWeight.set || GScale.set , IWeight.value , GScale.value ,
			!NoBlackOut.set , Progress.set
			) )
			fprintf(stderr,"SocketedMultigridServer::SetUp failed\n");
	}
	else
	{
		if( !server.SetUp( Prefix.value , Port.value , Count.value , Iters.value , InCoreRes.value , MinMGRes.value , VCycles.value , MinBandSize.value , TileWidth.value , TileHeight.value , TileExtension.value , Quality.value , HDR.set , Lanes.value , Verbose.set , Spherical.set , IWeight.set || GScale.set , IWeight.value , GScale.value ,
			!NoBlackOut.set , Progress.set
			) )
			fprintf(stderr,"SocketedMultigridServer::SetUp failed\n");
	}

	double t=Time();
	server.Run();
	size_t current,peak;
	WorkingSetInfo(current,peak);
	printf("Running Time: %f\n",Time()-t) , fflush( stdout );
	printf("Peak working set: %d MB\n",peak>>20) , fflush( stdout );
	printf("Packets Sent / Received: %lld / %lld\n",packetsSent,packetsReceived) , fflush( stdout );

	MyWinSock::UnLoad();

	return EXIT_SUCCESS;
}
