/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#ifndef ARRAY_INCLUDED
#define ARRAY_INCLUDED

#define ASSERT_ARRAY_ACCESS 0
#define ASSERT( x ) { if( !( x ) ) __debugbreak(); }

#if ASSERT_ARRAY_ACCESS
#include "Array.inl"
#define      Pointer( ... )      Array< __VA_ARGS__ >
#define ConstPointer( ... ) ConstArray< __VA_ARGS__ >
template< class C > void        VFreeArray( Array< C >& a )                                     { a.Free( ); }
template< class C > void         FreeArray( Array< C >& a )                                     { a.Free( ); }
template< class C > Array< C > VAllocArray( int size , int alignment=1 , const char* name=NULL ){ return Array< C >( size , alignment , false , true  , name ); }
template< class C > Array< C >  AllocArray( int size , int alignment=1 , const char* name=NULL ){ return Array< C >( size , alignment , false , false , name ); }
template< class C > Array< C > NullPointer( void )                                              { return Array< C >( ); }
template< class C > C*      PointerAddress( Array< C >& a )                                     { return a.pointer(); }
template< class C > Array< C >  GetPointer( C& c )                                              { return Array< C >::FromPointer( &c , 1 ); }
#else // !ASSERT_ARRAY_ACCESS
#define      Pointer( ... )       __VA_ARGS__*
#define ConstPointer( ... ) const __VA_ARGS__*
#ifdef WIN32
template< class C > C*     AllocArray( int size , int alignment=1 , const char* name=NULL ){ return (C*)_aligned_malloc( sizeof(C) * size , alignment ); }
#define                     FreeArray( ptr ) if( ptr ) _aligned_free( ptr ) , ptr = NULL
#else // !WIN32
template< class C > C*     AllocArray( int size , int alignment=1 , const char* name=NULL ){ return (C*)memalign( sizeof(C) * size , alignment ); }
#define                     FreeArray( ptr ) if( ptr ) free( ptr ) , ptr = NULL
#endif // WIN32
#define                    VFreeArray( ptr ) if( ptr ) VirtualFree( ptr , 0 , MEM_RELEASE ) , ptr = NULL
template< class C > C*    VAllocArray( int size , int alignment=1 , const char* name=NULL ){ return (C*)VirtualAlloc( NULL , size*sizeof(C) , MEM_RESERVE | MEM_COMMIT , PAGE_READWRITE ); }
template< class C > C*    NullPointer( void ){ return NULL; }
template< class C > C* PointerAddress( C* c ){ return c; }
template< class C > C*     GetPointer( C& c ){ return &c; }
#endif // ASSERT_ARRAY_ACCESS
#endif // ARRAY_INCLUDED
