/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#ifndef MEMORY_STREAM_INCLUDED
#define MEMORY_STREAM_INCLUDED


#include <windows.h>

template<class Data>
class DataStream
{
public:
	virtual ~DataStream(void){;};
	virtual Data& operator [] (size_t idx) = 0;
	virtual size_t size(void) const = 0;
	virtual void reset(void) = 0;
};
template<class Data>
class MemoryStream : public DataStream<Data>
{
	bool deleteData;
	Pointer( Data ) data;
	size_t _size;
public:
	MemoryStream(size_t size,int byteAlignment=1,bool zeroMemory=true);
	MemoryStream( Pointer( Dat ) data , size_t size );
	~MemoryStream(void);
	Data& operator[] (size_t idx);
	size_t size(void) const;
	void reset(void);
};
template<class Data>
class FileMappedStream : public DataStream<Data>
{
	HANDLE hFile;
	HANDLE data;
	size_t _size,_blockSize;
	Data *baseData,*_baseData,*dataPtr;
public:
	FileMappedStream(const char* fileName,size_t blockSize,int byteAlignment=1);
	FileMappedStream(size_t size,size_t blockSize,int byteAlignment=1);
	~FileMappedStream(void);
	Data& operator[] (size_t idx);
	size_t size(void) const;
	void reset(void);
};
#include "DataStream.inl"
#endif // MEMORY_STREAM_INCLUDED