/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
//////////////////
// MemoryStream //
//////////////////
template<class Data>
MemoryStream<Data>::MemoryStream(size_t size,int byteAlignment,bool zeroMemory)
{
	data = AllocArray< Data >( size , byteAlignment , "MemoryStream< Data >::data" );
	if( zeroMemory ) memset( data , 0 , sizeof( Data ) * size );
	_size=size;
	deleteData=true;
}
template<class Data>
MemoryStream<Data>::MemoryStream( Pointer( Data ) data,size_t size)
{
	this->data = data;
	_size=size;
	deleteData=false;
}
template<class Data>
MemoryStream<Data>::~MemoryStream(void)
{
	if(deleteData && data)	FreeArray( data );
	data = NullPointer< NULL >( );
	_size=0;
}
template<class Data>
inline Data& MemoryStream<Data>::operator [] (size_t idx)
{
	return *(data+idx);
}
template<class Data>
inline size_t MemoryStream<Data>::size(void) const	{ return _size; }
template<class Data>
inline void MemoryStream<Data>::reset(void){ ; }


//////////////////////
// FileMappedStream //
//////////////////////
#include <cstringt.h>
#include <atlstr.h>


template<class Data>
FileMappedStream<Data>::FileMappedStream(const char* fileName,size_t blockSize,int byteAlignment)
{
	CString s(fileName);
	hFile = CreateFile(s,FILE_ALL_ACCESS,0,NULL,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
	data  = CreateFileMapping(hFile, NULL, PAGE_READWRITE | SEC_COMMIT, 0, 0, NULL);
	_baseData = (Data*)MapViewOfFile(data,FILE_MAP_ALL_ACCESS,0,0,0);
	baseData  = (Data*)(((size_t)(_baseData)+byteAlignment-1) & ~(byteAlignment-1));
	_size=0;
	_blockSize=blockSize;
}

template<class Data>
FileMappedStream<Data>::FileMappedStream(size_t size,size_t blockSize,int byteAlignment)
{
	CString s(tmpnam(NULL));
	hFile = CreateFile(s,FILE_ALL_ACCESS,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL | FILE_FLAG_DELETE_ON_CLOSE | FILE_ATTRIBUTE_TEMPORARY,NULL);
	data  = CreateFileMapping(hFile, NULL, PAGE_READWRITE | SEC_COMMIT, 0, size*sizeof(Data)+byteAlignment-1, NULL);
	_baseData = (Data*)MapViewOfFile(data,FILE_MAP_ALL_ACCESS,0,0,size*sizeof(Data)+byteAlignment-1);
	baseData  = (Data*)(((size_t)(_baseData)+byteAlignment-1) & ~(byteAlignment-1));
	_size=size;
	_blockSize=blockSize;
}
template<class Data>
FileMappedStream<Data>::~FileMappedStream(void)
{
	UnmapViewOfFile(_baseData);
	CloseHandle(hFile);
}
template<class Data>
void FileMappedStream<Data>::reset(void)
{
	dataPtr = baseData;
}
template<class Data>
size_t FileMappedStream<Data>::size(void) const
{
	return _size;
}
template<class Data>
Data& FileMappedStream<Data>::operator [] (size_t idx)
{
	dataPtr=baseData+idx;
	return *dataPtr;
}

