/*
Copyright (c) 2008, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include "JPEGStream.h"

template<> inline		bool IsFloatType<half>		(void)	{	return true;	}
template<> inline		bool IsFloatType<float>		(void)	{	return true;	}
template<> inline		bool IsFloatType<double>	(void)	{	return true;	}
template<class Real>	bool IsFloatType			(void)	{	return false;	}

template<class Real>
class ConversionFactor
{
public:
	static const double Factor;
};
template <> const double ConversionFactor<char>::Factor				= double((long long(1)<< 8)-1);
template <> const double ConversionFactor<unsigned char>::Factor	= double((long long(1)<< 8)-1);
template <> const double ConversionFactor<__int16>::Factor			= double((long long(1)<<16)-1);
template <> const double ConversionFactor<unsigned __int16>::Factor = double((long long(1)<<16)-1);
template <> const double ConversionFactor<int>::Factor				= double((long long(1)<<32)-1);
template <> const double ConversionFactor<unsigned int>::Factor		= double((long long(1)<<32)-1);
template<class Real>	const double ConversionFactor<Real>::Factor	= 1.0;

template<class In,class Out>
class Converter
{
public:
	static const double ConvertFactor;
};
template<class In,class Out>
const double Converter<In,Out>::ConvertFactor = ConversionFactor<Out>::Factor / ConversionFactor<In>::Factor;

template<class In,class Out>
inline Out ConvertChannel(const In& in)
{
	return Out(double(in)*Converter<In,Out>::ConvertFactor);
}

//////////////////////
// WriteImageStream //
//////////////////////
template< class PReal , class BReal , int Channels >
WriteImageStream< PReal , BReal , Channels >::WriteImageStream(void* (*IW)(char*,int,int,int,MultiStreamIOServer*),void (*WR)(void*,void*,int),void (*FW)(void*),					
															   char* fileName,int width,int height,bool separateColors,int quality,bool clamp,
															   MultiStreamIOServer* ioServer ) 
{
	_InitWrite		= IW;
	_WriteRow		= WR;
	_FinalizeWrite	= FW;
	current=0;
	_q=quality;
	_w=width;
	_h=height;
	_clamp=clamp;
	_separate=separateColors;
	info=NULL;
	pixels = NullPointer< PReal >( );
	buffer = NullPointer< BReal >( );
	Init( fileName , ioServer );
}
template< class PReal , class BReal , int Channels >
void WriteImageStream< PReal , BReal , Channels >::Init( char* fileName , MultiStreamIOServer* ioServer )
{
	memset( average , 0 , sizeof( average ) );

	FreeArray( pixels );
	FreeArray( buffer );
	pixels = AllocArray< PReal >( _w * Channels , 1 , "WriteImageStream::pixels" );
	buffer = AllocArray< BReal >( _w * Channels , 1 , "WriteImageStream::buffer" );
	if(!buffer)	fprintf(stderr,"Failed to allocate buffer\n") , exit(0);
	if(!pixels)	fprintf(stderr,"Failed to allocate pixels\n") , exit(0);
	info=_InitWrite(fileName,_w,_h,_q,ioServer);
}

template< class PReal , class BReal , int Channels >
WriteImageStream< PReal , BReal , Channels >::~WriteImageStream(void)
{
	FreeArray( buffer );
	FreeArray( pixels );
	_FinalizeWrite(info);
	for( int c=0 ; c<Channels ; c++ ) average[c] /= _w , average[c] /= _h;
}
template< class PReal , class BReal , int Channels > int  WriteImageStream< PReal , BReal , Channels >::rows(void) const {return _h;}
template< class PReal , class BReal , int Channels > int  WriteImageStream< PReal , BReal , Channels >::rowSize(void) const {return _w*Channels*sizeof(BReal);}
template< class PReal , class BReal , int Channels > bool WriteImageStream< PReal , BReal , Channels >::SeparateColors(void)	const	{return _separate;}
template< class PReal , class BReal , int Channels >
Pointer( byte ) WriteImageStream< PReal , BReal , Channels >::operator[] (int idx)
{
	return ( Pointer( byte ) )buffer;
}
template< class PReal , class BReal , int Channels >
void WriteImageStream< PReal , BReal , Channels >::advance(void)
{
	if(current<_h)
	{
		if(!_clamp)
			if(_separate)
			{
				for(int x=0;x<_w;x++)
				{
					for( int c=0 ; c<Channels ; c++ )
					{
						pixels[x*Channels+c]=ConvertChannel< BReal , PReal >( buffer[ x+c*_w] );
						average[c] += pixels[x*Channels+c];
					}
				}
			}
			else
			{
				for(int x=0;x<_w;x++)
				{
					for( int c=0 ; c<Channels ; c++ )
					{
						pixels[x*Channels+c] = ConvertChannel<BReal,PReal>( buffer[x*Channels+c] );
						average[c] += pixels[x*Channels+c];
					}
				}
			}
		else
			for(int x=0;x<_w;x++)
			{
				double _values[ Channels ];

				if( _separate ) for( int c=0 ; c<Channels ; c++ ) _values[c] =ConvertChannel<BReal,double>(buffer[x+c*_w]);
				else            for( int c=0 ; c<Channels ; c++ ) _values[c] =ConvertChannel<BReal,double>(buffer[x*Channels+c]);
				for( int c=0 ; c<Channels ; c++ )
				{
					if		(_values[c]<0)	_values[c]=0;
					else if	(_values[c]>1)	_values[c]=1;
					pixels[x*Channels+c]=ConvertChannel<double,PReal>(_values[c]);
					average[c] += pixels[x*Channels+c];
				}
			}
		_WriteRow( PointerAddress( pixels ) , info , current );
	}
	current++;
}
/////////////////////
// ReadImageStream //
/////////////////////
template< class PReal , class BReal , int Channels >
ReadImageStream< PReal , BReal , Channels >::ReadImageStream(void* (*IR)(char*,int&,int&,MultiStreamIOServer*),void (*RR)(void*,void*,int),void (*FR)(void*),					
															 char* fileName,int& width,int& height,bool separateColors ,
															 MultiStreamIOServer* ioServer )
{
	_InitRead		= IR;
	_ReadRow		= RR;
	_FinalizeRead	= FR;

	_separate		= separateColors;
	current=-1;

	info=_InitRead(fileName,width,height , ioServer );
	_w=width;
	_h=height;

	pixels = AllocArray< PReal >( _w * Channels , 1 , "ReadImageStream::pixels" );
	buffer = AllocArray< BReal >( _w * Channels , 1 , "ReadImageStream::buffer" );

	if(!buffer)	fprintf(stderr,"Failed to allocate buffer\n") , exit(0);
	if(!pixels)	fprintf(stderr,"Failed to allocate pixels\n") , exit(0);

	advance();
}
template< class PReal , class BReal , int Channels >
ReadImageStream< PReal , BReal , Channels >::~ReadImageStream(void)
{
	FreeArray( pixels );
	FreeArray( buffer );
	_FinalizeRead(info);
	for( int c=0 ; c<Channels ; c++ ) average[c] /= _w , average[c] /= _h;
}
template< class PReal , class BReal , int Channels >	int  ReadImageStream< PReal , BReal , Channels >::rows(void) const {return _h;}
template< class PReal , class BReal , int Channels >	int  ReadImageStream< PReal , BReal , Channels >::rowSize(void) const {	return _w*Channels*sizeof(BReal); }
template< class PReal , class BReal , int Channels >	bool ReadImageStream< PReal , BReal , Channels >::SeparateColors(void) const { return _separate; }
template< class PReal , class BReal , int Channels >
Pointer( byte ) ReadImageStream< PReal , BReal , Channels >::operator[] (int idx)
{
	return ( Pointer( byte ) ) buffer;
}
template< class PReal , class BReal , int Channels >
void ReadImageStream< PReal , BReal , Channels >::advance(void)
{
	current++;
	if(current<_h)
	{
		_ReadRow( PointerAddress( pixels ) , info , current );
		if(_separate)
			for(int x=0;x<_w;x++)
			{
				for( int c=0 ; c<Channels ; c++ )
				{
					average[c] += double(pixels[x*Channels+c]);
					buffer[x+c*_w] = ConvertChannel<PReal,BReal>( pixels[x*Channels+c]);
				}
			}
		else
			for(int x=0;x<_w;x++)
			{
				for( int c=0 ; c<Channels ; c++ )
				{
					average[c]+=double(pixels[x*Channels+c]);
					buffer[x*Channels+c]=ConvertChannel<PReal,BReal>(pixels[x*Channels+c]);
				}
			}
	}
}

//////////////////
// WImageStream //
//////////////////
template<class Real,class OutType>
WImageStream<Real,OutType>::WImageStream(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(InitWrite<OutType>,WriteRow<OutType>,FinalizeWrite<OutType>,fileName,width,height,separateColors,quality,false , ioServer )
{
}
/////////////////////
// BMPWImageStream //
/////////////////////
template<class Real>
BMPWImageStream<Real>::BMPWImageStream(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(BMPInitWrite,BMPWriteRow,BMPFinalizeWrite,fileName,width,height,separateColors,quality,true , ioServer )
{
}
/////////////////////
// JPEGWImageSteam //
/////////////////////
template<class Real>
JPEGWImageStream< Real >::JPEGWImageStream( char* fileName , int width , int height , bool separateColors , int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(JPEGInitWriteColor,JPEGWriteRow,JPEGFinalizeWrite,fileName,width,height,separateColors,quality,true , ioServer )
{
}
template< class Real >
int JPEGWImageStream< Real >::Service( void )
{
	JPEGWriteInfo* inf = (JPEGWriteInfo*) info;
	return inf->outStream->Service();
}
////////////////////
// PNGWImageSteam //
////////////////////
template<class Real>
PNGWImageStream<Real>::PNGWImageStream(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(PNGInitWriteColor,PNGWriteRow,PNGFinalizeWrite,fileName,width,height,separateColors,quality,true , ioServer )
{
}
/////////////////////
// TIFFWImageSteam //
/////////////////////
template<class Real>
TIFFWImageStream<Real>::TIFFWImageStream(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(TIFFInitWriteColor,TIFFWriteRow,TIFFFinalizeWrite,fileName,width,height,separateColors,quality,true , ioServer )
{
}
///////////////////////
// PNGWImageSteamHDR //
///////////////////////
template<class Real>
PNGWImageStreamHDR<Real>::PNGWImageStreamHDR(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(PNGInitWriteColorHDR,PNGWriteRowHDR,PNGFinalizeWrite,fileName,width,height,separateColors,quality,true , ioServer )
{
}
////////////////////////
// TIFFWImageSteamHDR //
////////////////////////
template<class Real>
TIFFWImageStreamHDR<Real>::TIFFWImageStreamHDR(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(TIFFInitWriteColorHDR,TIFFWriteRowHDR,TIFFFinalizeWrite,fileName,width,height,separateColors,quality,true , ioServer )
{
}
/////////////////////
// WDPWImageStream //
/////////////////////
template<class Real>
WDPWImageStream<Real>::WDPWImageStream(char* fileName,int width,int height,bool separateColors,int quality , MultiStreamIOServer* ioServer ) :
WriteImageStream(WDPInitWriteColor,WDPWriteColorRow,WDPFinalizeWrite,fileName,width,height,separateColors,quality,false , ioServer )
{
}
//////////////////
// RImageStream //
//////////////////
template<class Real,class InType>
RImageStream<Real,InType>::RImageStream(char* fileName,int& width,int& height,bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream(InitRead<InType>,ReadRow<InType>,FinalizeRead<InType>,fileName,width,height,separateColors , ioServer )
{
}
//////////////////////
// JPEGRImageStream //
//////////////////////
template<class Real>
JPEGRImageStream<Real>::JPEGRImageStream(char* fileName,int& width,int& height,bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream(JPEGInitReadColor,JPEGReadRow,JPEGFinalizeRead,fileName,width,height,separateColors , ioServer )
{
}
template< class Real >
int JPEGRImageStream< Real >::Service( void )
{
	JPEGReadInfo* inf = (JPEGReadInfo*) info;
	return inf->inStream->Service();
}
/////////////////////
// BMPRImageStream //
/////////////////////
template< class Real >
BMPRImageStream< Real >::BMPRImageStream( char* fileName , int& width , int& height , bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream( BMPInitReadColor , BMPReadColorRow , BMPFinalizeRead , fileName , width , height , separateColors , ioServer )
{
}
/////////////////////
// PNGRImageStream //
/////////////////////
template< class Real >
PNGRImageStream< Real >::PNGRImageStream( char* fileName , int& width , int& height , bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream( PNGInitReadColor , PNGReadColorRow , PNGFinalizeRead , fileName , width , height , separateColors , ioServer )
{
}
//////////////////////////
// PNGRImageStreamAlpha //
//////////////////////////
template< class Real >
PNGRImageStreamAlpha< Real >::PNGRImageStreamAlpha( char* fileName , int& width , int& height , bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream( PNGInitReadColorAlpha , PNGReadColorRow , PNGFinalizeRead , fileName , width , height , separateColors , ioServer )
{
}

/////////////////////
// WDPRImageStream //
/////////////////////
template<class Real>
WDPRImageStream<Real>::WDPRImageStream(char* fileName,int& width,int& height,bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream(WDPInitReadColor,WDPReadColorRow,WDPFinalizeRead,fileName,width,height,separateColors , ioServer )
{
}
//////////////////////
// PFMRImageStream //
//////////////////////
template<class Real>
PFMRImageStream<Real>::PFMRImageStream( char* fileName , int& width , int& height , bool separateColors , MultiStreamIOServer* ioServer ) :
ReadImageStream( PFMInitReadColor , PFMReadRow , PFMFinalizeRead , fileName , width , height , separateColors , ioServer )
{
}

////////////////////////
// StreamingInputTile //
////////////////////////
template<class Real>
StreamingInputTile<Real>::StreamingInputTile(void)
{
	rc=1;
	data=NULL;
	w=h=0;
	sX=sY=0;
	rows=NULL;
}
template<class Real>
StreamingInputTile<Real>::~StreamingInputTile(void)
{
	if(data)	delete data;
	if(rows)
	{
		for(int i=0;i<rc;i++)	if(rows[i])	delete[] rows[i];
		delete[] rows;
	}
	data=NULL;
	rows=NULL;
	rc=0;
};

template<class Real>
void StreamingInputTile<Real>::Init( const char* tName , int rowCount , int startX , int startY , bool separateColors = false )
{
	average[0]=average[1]=average[2]=0;
	strcpy(tileName,tName);
	rc=rowCount;
	sX=startX;
	sY=startY;
	if(rowCount<1)	exit(0);
	rows = new Real*[rc];
	data = GetReadStream< Real >( tileName , w , h , separateColors , NULL );
	for(int i=-rc;i<0;i++)	init(i);
}

template<class Real>
void StreamingInputTile<Real>::init( int idx )
{
	if( idx+rc-1==sY )
	{
		if(!data)	exit(0);
		for(int i=0;i<rc;i++)	rows[i] = new Real[w*3];
		for(int i=0;i<rc-1;i++)
		{
			memcpy( rows[i] , (*data)[i] , w*sizeof(Real)*3 );
			data->advance();
		}
	}
	else if(idx==sY+h)
	{
		delete data;
		for(int i=0;i<rc;i++) delete[] rows[i];
		delete[] rows;
		data=NULL;
		rows=NULL;
	}
	if(idx>=sY && data)
	{
		memcpy( rows[(idx+rc-1-sY)%rc] , (*data)[idx+rc-1-sY] , w*3*sizeof(Real) );
		data->advance();
	}
}
template<class Real>
Color<Real> StreamingInputTile<Real>::operator() ( int i , int j )
{
	Color<Real> clr;
	if( this->data->SeparateColors( ) )
	{
		Real* data = &rows[(j-sY)%rc][i-sX];
		clr[0] = data[0*w];
		clr[1] = data[1*w];
		clr[2] = data[2*w];
	}
	else
	{
		Real* data = &rows[(j-sY)%rc][3*(i-sX)];
		clr[0] = data[0];
		clr[1] = data[1];
		clr[2] = data[2];
	}
	return clr;
}
template<class Real>
Color<Real> StreamingInputTile<Real>::operator() ( int i , int j , Real& a )
{
	Color<Real> clr;
	if( this->data->SeparateColors() )
	{
		Real* data=&rows[(j-sY)%rc][i-sX];
		clr[0] = data[0*w];
		clr[1] = data[1*w];
		clr[2] = data[2*w];
	}
	{
		Real* data=&rows[(j-sY)%rc][3*(i-sX)];
		clr[0] = data[0];
		clr[1] = data[1];
		clr[2] = data[2];
	}
	return clr;
}
template<class Real>
int StreamingInputTile<Real>::width(void)	const	{return w;}
template<class Real>
int StreamingInputTile<Real>::height(void)	const	{return h;}
template<class Real>
int StreamingInputTile<Real>::startX(void) const	{return sX;}
template<class Real>
int StreamingInputTile<Real>::startY(void) const	{return sY;}


///////////////////
// RGridOfImages //
///////////////////
template< class Real >
void RGridOfImages< Real >::GetImageSize( char* fileName , int& width , int& height )
{
	int c , r;
	FILE* fp = fopen(fileName,"r");
	char imageName[1024];
	if( fscanf( fp , "Columns: %d " , &c ) != 1) fprintf( stderr , "Failed to read in columns in RGridOfImages\n" ) , exit( 0 );
	if( fscanf( fp , "Rows: %d " , &r )     !=1) fprintf( stderr , "Failed to read in rows in RGridOfImages\n" )    , exit( 0 );
	width = height = 0;
	int* widths  = new int[c];
	int* heights = new int[r];
	for ( int j = 0 ; j < r ; j++ )
		for ( int i = 0 ; i < c ; i++ )
		{
			if( fscanf( fp , " %s " , imageName )!=1 )	fprintf(stderr,"Failed to read in image [%d][%d] in RGridOfImages\n",i,j) , exit(0);
			int ww , hh;
			GetReadSize< Real >( imageName , ww , hh );
			if( !j ) widths[i]  = ww	,	width += ww;
			if( !i ) heights[j] = hh	,	height += hh;
			if( widths[i]!=ww )		fprintf( stderr , "Inconsistent widths in column [%d]: %d != %d\n" , i ,  widths[i] , ww );
			if( heights[j]!=hh )	fprintf( stderr , "Inconsistent heights in row [%d]: %d != %d\n"   , j , heights[j] , hh );
		}
	fclose(fp);
	delete[] widths;
	delete[] heights;

}
template<class Real>
RGridOfImages<Real>::RGridOfImages(char* fileName,int& width,int& height,bool separateColors , MultiStreamIOServer* ioServer)
{
	FILE* fp = fopen(fileName,"r");
	char imageName[1024];
	if(fscanf(fp,"Columns: %d ",&_c) != 1)	fprintf(stderr,"Failed to read in columns in RGridOfImages\n") , exit(0);
	if(fscanf(fp,"Rows: %d ",&_r) !=1)		fprintf(stderr,"Failed to read in rows in RGridOfImages\n")    , exit(0);

	int w,h;
	_separate=separateColors;
	_widths  = new int[_c];
	_heights = new int[_r];
	_grid = new StreamingGrid**[_c];
	_fileNames = new char**[_c];
	for ( int i = 0 ; i < _c ; i++ )
	{
		_grid[i] = new StreamingGrid*[_r];
		_fileNames[i] = new char*[_r];
	}
	_w = _h = 0;
	for ( int j = 0 ; j < _r ; j++ )
		for ( int i = 0 ; i < _c ; i++ )
		{
			if(fscanf(fp," %s ",imageName)!=1)	fprintf(stderr,"Failed to read in image [%d][%d] in RGridOfImages\n",i,j) , exit(0);
			_fileNames[i][j]=new char[strlen(imageName)+1];
			strcpy( _fileNames[i][j] , imageName );
			GetReadSize< Real >( _fileNames[i][j] , w , h );
			if( !j ) _widths[i]  = w	,	_w += w;
			if( !i ) _heights[j] = h	,	_h += h;
			if( _widths[i]!=w )  fprintf( stderr , "Inconsistent widths in column [%d]: %d != %d\n" , i , _widths[i]  , w ); 
			if( _heights[j]!=h ) fprintf( stderr , "Inconsistent heights in row   [%d]: %d != %d\n" , j , _heights[j] , h );
		}
	fclose(fp);
	_buffer = AllocArray< Real >( _w * 3 , 1 , "RGridOfImages::_buffer" );
	_current = -1;
	advance();
	width = _w;
	height = _h;
	_server = ioServer;
}
template<class Real>
RGridOfImages<Real>::~RGridOfImages(void)
{
	delete[] _widths;
	delete[] _heights;
	_widths = _heights = NULL;
	for (int i = 0 ; i < _c ; i++ )
	{
		for ( int j = 0 ; j <_r ; j++ )
		{
			if(_grid[i][j])			delete _grid[i][j];
			if(_fileNames[i][j])	delete _fileNames[i][j];
		}
		delete[] _grid[i];
		delete[] _fileNames[i];
	}
	delete[] _grid;
	delete[] _fileNames;
	FreeArray( _buffer );
	_grid = NULL;
	_fileNames = NULL;
}
template<class Real>	int RGridOfImages<Real>::rows(void)		const	{	return _h;	}
template<class Real>	int RGridOfImages<Real>::rowSize(void)	const	{	return _w*3*sizeof(Real);	}
template<class Real>	bool RGridOfImages<Real>::SeparateColors(void)	const	{	return _separate;	}
template<class Real>
Pointer( byte ) RGridOfImages<Real>::operator [] (int idx)
{
	return ( Pointer( byte ) )( _buffer );
}
template<class Real>
void RGridOfImages<Real>::advance(void)
{
	_current++;
	if(_current<_h)
	{
		int row;
		int x = 0 , y = _current;
		for ( row = 0 ; y >= _heights[row] ; y -= _heights[row++] ) ;

		for( int col = 0 ; col < _c ; col++ )
		{
			if( y == 0 )
			{
				int w,h;
				_grid[col][row] = GetReadStream<Real>( _fileNames[col][row] , w , h , _separate , _server );
				_grid[col][row]->SetServer( _server );
			}
			if( !_grid[col][row] )	fprintf(stderr,"Error: Attempting to read from NULL tile\n") , exit(0);
			Pointer( Real ) subRow = ( Pointer( Real ) )(*_grid[col][row])[y];

			if(_separate)
			{
				memcpy(_buffer+x+0*_w,subRow+0*_widths[col],sizeof(Real)*_widths[col]);
				memcpy(_buffer+x+1*_w,subRow+1*_widths[col],sizeof(Real)*_widths[col]);
				memcpy(_buffer+x+2*_w,subRow+2*_widths[col],sizeof(Real)*_widths[col]);
			}
			else
				memcpy(_buffer+3*x,subRow,sizeof(Real)*_widths[col]*3);

			x+=_widths[col];
			_grid[col][row]->advance();

			if( y >= _grid[col][row]->rows()-1 )
			{
				delete _grid[col][row];
				_grid[col][row]=NULL;
			}
		}
	}
}
///////////////////
// WGridOfImages //
///////////////////
template<class Real>
WGridOfImages<Real>::WGridOfImages(char* fileName,int width,int height,bool separateColors,int quality,int tileWidth,int tileHeight , MultiStreamIOServer* ioServer , bool hdr )
{
	char* header = GetFileHeader( fileName );
	_init( fileName , header , DefaultOutputTileExtension , width , height , separateColors , quality , tileWidth , tileHeight , ioServer , hdr );
	delete[] header;
}
template<class Real>
WGridOfImages<Real>::WGridOfImages(char* fileName,const char* header,const char* ext,int width,int height,bool separateColors,int quality,int tileWidth,int tileHeight , MultiStreamIOServer* ioServer , bool hdr )
{
	_init(fileName,header,ext,width,height,separateColors,quality,tileWidth,tileHeight , ioServer , hdr );
}
template<class Real>
void WGridOfImages<Real>::_init(char* fileName,const char* header,const char* ext,int width,int height,bool separateColors,int quality,int tileWidth,int tileHeight , MultiStreamIOServer* ioServer , bool hdr )
{
	FILE* fp = fopen(fileName,"w");
	_separate=separateColors;
	_w = width;
	_h = height;
	_c = (width + tileWidth -1) / tileWidth;
	_r = (height + tileHeight -1) / tileHeight;
	_quality = quality;
	_hdr = hdr;
	char imageName[1024];
	fprintf(fp,"Columns: %d\n",_c) , fflush( fp );
	fprintf(fp,"Rows: %d\n",_r)    , fflush( fp );

	int w,h;
	_widths  = new int[_c];
	_heights = new int[_r];
	for ( int i = 0 ; i < _c ; i++ )	_widths[i]  = tileWidth;
	for ( int j = 0 ; j < _r ; j++ )	_heights[j] = tileHeight;
	_widths[_c-1]  = width  - (_c-1)*tileWidth;
	_heights[_r-1] = height - (_r-1)*tileHeight;

	_grid = new StreamingGrid**[_c];
	_fileNames = new char**[_c];
	for ( int i = 0 ; i < _c ; i++ )
	{
		_grid[i] = new StreamingGrid*[_r];
		_fileNames[i] = new char*[_r];
	}
	for ( int j = 0 ; j < _r ; j++ )
		for ( int i = 0 ; i < _c ; i++ )
		{
			sprintf( imageName , "%s.%d.%d.%s" , header , i , j , ext );
			_fileNames[i][j]=new char[strlen(imageName)+1];
			strcpy(_fileNames[i][j],imageName);
			fprintf(fp,"%s\n",imageName) , fflush( fp );
			_grid[i][j] = NULL;
		}
	fclose(fp);
	_buffer = AllocArray< Real >( _w * 3 , 1 , "WGridOfImages::_buffer" );
	_current = 0;
	_server = ioServer;
}
template<class Real>
WGridOfImages<Real>::~WGridOfImages(void)
{
	delete[] _widths;
	delete[] _heights;
	_widths = _heights = NULL;
	for (int i = 0 ; i < _c ; i++ )
	{
		for ( int j = 0 ; j <_r ; j++ )
		{
			if(_grid[i][j])			delete _grid[i][j];
			if(_fileNames[i][j])	delete _fileNames[i][j];
		}
		delete[] _grid[i];
	}
	delete[] _grid;
	FreeArray( _buffer );
	_grid = NULL;
	_fileNames = NULL;
}
template<class Real>	int WGridOfImages<Real>::rows(void)		const	{	return _h;	}
template<class Real>	int WGridOfImages<Real>::rowSize(void)	const	{	return _w*3*sizeof(Real);	}
template<class Real>	bool WGridOfImages<Real>::SeparateColors(void)	const	{	return _separate;	}
template<class Real>
Pointer( byte ) WGridOfImages<Real>::operator [] (int idx)
{
	return ( Pointer( byte ) )( _buffer );
}

template<class Real>
void WGridOfImages<Real>::advance(void)
{
	if(_current<_h)
	{
		int row;
		int x = 0 , y = _current;
		for ( row = 0 ; y >= _heights[row] ; y -= _heights[row++] ) ;

		for( int col = 0 ; col < _c ; col++ )
		{
			if ( y==0 )
			{
				_grid[col][row] = GetWriteStream<Real>( _fileNames[col][row] , _widths[col] , _heights[row] , _separate , _quality , _server , _hdr );
			}
			if ( !_grid[col][row])	fprintf(stderr,"Error: Attempting to write to NULL tile\n") , exit(0);
			Pointer( Real ) subRow = ( Pointer( Real ) )(*_grid[col][row])[y];
			if(_separate)
			{
				memcpy(subRow+0*_widths[col],_buffer+x+0*_w,sizeof(Real)*_widths[col]);
				memcpy(subRow+1*_widths[col],_buffer+x+1*_w,sizeof(Real)*_widths[col]);
				memcpy(subRow+2*_widths[col],_buffer+x+2*_w,sizeof(Real)*_widths[col]);
			}
			else
				memcpy(subRow,_buffer+3*x,sizeof(Real)*_widths[col]*3);

			x+=_widths[col];
			_grid[col][row]->advance();

			if ( y>=_grid[col][row]->rows()-1 )
			{
				delete _grid[col][row];
				_grid[col][row] = NULL;
			}
		}
	}
	_current++;
}



/////////////////////////////////////////////
// MultiThreadedRGridOfImages::ImageReader //
/////////////////////////////////////////////
template< class Real >
void MultiThreadedRGridOfImages< Real >::ImageReader::init( int columns , int rows ,
														    int* widths , int* heights ,
															char*** fileNames ,
															bool separateColors ,
															Pointer( Real ) buffer ,
															HANDLE loadHandle , HANDLE unloadHandle ,
															MultiStreamIOServer* ioServer )
{
if( separateColors )
{
	fprintf( stderr , "MultiThreadedRGridOfImages does not support color separation\n" );
	exit( 0 );
}

	_c				= columns;
	_r				= rows;
	_widths			= widths;
	_heights		= heights;
	_separate		= separateColors;
	_fileNames		= fileNames;
	_buffer			= buffer;
	_loadHandle		= loadHandle;
	_unloadHandle	= unloadHandle;
	_server = ioServer;
	_w = _h = 0;
	for( int i=0 ; i<_c ; i++ ) _w += _widths[i];
	for( int i=0 ; i<_r ; i++ ) _h += _heights[i];
}
template< class Real >
void MultiThreadedRGridOfImages< Real >::ImageReader::Run( void )
{
	StreamingGrid** grids = new StreamingGrid*[_c];
	for( int i=0 ; i<_h ; i++ )
	{
		MyWaitForSingleObject( _loadHandle , 10000 , "MultiThreadedRGridOfImages<Real>::ImageReader::Run" );
		int row , x = 0 , y = i;
		for ( row = 0 ; y >= _heights[row] ; y -= _heights[row++] ) ;
		for( int col = 0 ; col < _c ; col++ )
		{
			if( y == 0 )
			{
				int w,h;
				grids[col] = GetReadStream< Real >(_fileNames[col][row] , w , h , _separate , _server );
			}
			if( !grids[col] )	fprintf(stderr,"Error: Attempting to read from NULL tile\n") , exit(0);
			Pointer( Real ) subRow = ( Pointer( Real ) )(*grids[col])[y];
			memcpy( _buffer+3*x , subRow , sizeof(Real)*_widths[col]*3 );

			x += _widths[col];
			grids[col]->advance();

			if( y >= grids[col]->rows()-1 )
			{
				delete grids[col];
				grids[col]=NULL;
			}
		}
		if( !SetEvent( _unloadHandle ) ) fprintfId( stderr , "Set event failed: " ) , PrintError();
	}
	delete[] grids;
}

////////////////////////////////
// MultiThreadedRGridOfImages //
////////////////////////////////
template< class Real >
void MultiThreadedRGridOfImages< Real >::GetImageSize( char* fileName , int& width , int& height )
{
	RGridOfImages< Real >::GetImageSize( fileName , width , height );
}
template<class Real>
MultiThreadedRGridOfImages<Real>::MultiThreadedRGridOfImages( char* fileName , int& width , int& height , bool separateColors , int threads , int threadPriority , MultiStreamIOServer* ioServer )
{
	FILE* fp = fopen( fileName , "r" );
	char imageName[1024];
	if( fscanf( fp , "Columns: %d " , &_c ) !=1 )	fprintf( stderr , "Failed to read in columns in RGridOfImages\n" ) , exit(0);
	if( fscanf( fp , "Rows: %d " , &_r ) !=1 )		fprintf( stderr , "Failed to read in rows in RGridOfImages\n" )    , exit(0);

	int w,h;
	_threads	= threads;
	_separate	= separateColors;
	_widths		= new int[_c];
	_heights	= new int[_r];
	_fileNames	= new char**[_c];
	_dataReady	= false;
	if( _threads > _c ) _threads = _c;
	for ( int i = 0 ; i < _c ; i++ ) _fileNames[i] = new char*[_r];

	_w = _h = 0;
	for ( int j = 0 ; j < _r ; j++ )
		for ( int i = 0 ; i < _c ; i++ )
		{
			if( fscanf( fp , " %s " , imageName )!=1 ) fprintf( stderr , "Failed to read in image [%d][%d] in RGridOfImages\n" , i , j ) , exit(0);
			_fileNames[i][j]=new char[strlen(imageName)+1];
			strcpy( _fileNames[i][j] , imageName );
			GetReadSize< Real >( _fileNames[i][j] , w , h );
			if( !j ) _widths[i]  = w	,	_w += w;
			if( !i ) _heights[j] = h	,	_h += h;
			if( _widths[i]!=w )  fprintf( stderr , "Inconsistent widths in column [%d]: %d != %d\n" , i , _widths[i]  , w );
			if( _heights[j]!=h ) fprintf( stderr , "Inconsistent heights in row   [%d]: %d != %d\n" , j , _heights[j] , h );
		}
	fclose(fp);
	_buffer = AllocArray< Real >( _w * 3 , 1 , "MultiThreadedRGridOfImages::_buffer" );

	_imageReaderHandles	= new HANDLE[ _threads ];
	_loadHandles		= new HANDLE[ _threads ];
	_unloadHandles		= new HANDLE[ _threads ];
	_imageReaders		= new ImageReader[ _threads ];
	int offset = 0;
	for( int i=0 ; i<_threads ; i++ )
	{
		_loadHandles[i]		= CreateEvent( NULL , false , false , NULL );
		_unloadHandles[i]	= CreateEvent( NULL , false , true  , NULL );
		int start = (i*_c) / _threads;
		int end = ( (i+1)*_c ) / _threads;
		_imageReaders[i].init( end-start , _r , _widths+start , _heights , _fileNames + start , _separate , _buffer+offset*3 , _loadHandles[i] , _unloadHandles[i] , ioServer );
		for( int j=start ; j<end ; j++ ) offset += _widths[j];
		_imageReaderHandles[i] = SpawnThread< ImageReader >( _imageReaders+i , threadPriority );
	}
	_current = -1;
	advance();
	width	= _w;
	height	= _h;
}
template<class Real>
MultiThreadedRGridOfImages<Real>::~MultiThreadedRGridOfImages( void )
{
	delete[] _widths;
	delete[] _heights;
	_widths = _heights = NULL;
	for (int i = 0 ; i < _c ; i++ )
	{
		for ( int j = 0 ; j <_r ; j++ ) if(_fileNames[i][j])	delete _fileNames[i][j];
		delete[] _fileNames[i];
	}
	delete[] _fileNames;
	FreeArray( _buffer );
	_fileNames = NULL;
	MyWaitForMultipleObjects( _threads , _imageReaderHandles , 10000 , "MultiThreadedRGridOfImages<Real>::~MultiThreadedRGridOfImages" );
	for( int i=0 ; i<_threads ; i++ )
	{
		CloseHandle( _loadHandles[i] );
		CloseHandle( _unloadHandles[i] );
	}
	delete[] _imageReaders;
	delete[] _imageReaderHandles;
	delete[] _loadHandles;
	delete[] _unloadHandles;
}
template<class Real> int MultiThreadedRGridOfImages<Real>::rows(void)	 const	{ return _h;	}
template<class Real> int MultiThreadedRGridOfImages<Real>::rowSize(void) const	{ return _w*3*sizeof(Real);	}
template<class Real> bool MultiThreadedRGridOfImages<Real>::SeparateColors(void) const	{ return _separate;	}
template<class Real>
Pointer( byte ) MultiThreadedRGridOfImages<Real>::operator [] (int idx)
{
	if( !_dataReady )
		MyWaitForMultipleObjects( _threads , _unloadHandles , 10000 , "MultiThreadedRGridOfImages<Real>::operator[]" );
	_dataReady = true;
	return ( Pointer( byte ) )( _buffer );
}
template<class Real>
void MultiThreadedRGridOfImages<Real>::advance( void )
{
	_current++;
	if( _current<_h )
	{
		if( !_dataReady )
			MyWaitForMultipleObjects( _threads , _unloadHandles , 10000 , "MultiThreadedRGridOfImages<Real>::advance" );
		for( int i=0 ; i<_threads ; i++ ) if( !SetEvent( _loadHandles[i] ) ) fprintfId( stderr , "SetEvent failed: " ) , PrintError();
		_dataReady = false;
	}
}

/////////////////////////////////////////////
// MultiThreadedWGridOfImages::ImageWriter //
/////////////////////////////////////////////
template< class Real >
void MultiThreadedWGridOfImages< Real >::ImageWriter::init( int columns , int rows ,
														    int* widths , int* heights ,
															char*** fileNames ,
															bool separateColors , int quality ,
															Pointer( Real ) buffer ,
															HANDLE loadHandle , HANDLE unloadHandle ,
															MultiStreamIOServer* ioServer , bool hdr )
{
if( separateColors )
{
	fprintf( stderr , "MultiThreadedWGridOfImages does not support color separation\n" );
	exit( 0 );
}

	_c				= columns;
	_r				= rows;
	_widths			= widths;
	_heights		= heights;
	_separate		= separateColors;
	_quality		= quality;
	_fileNames		= fileNames;
	_buffer			= buffer;
	_loadHandle		= loadHandle;
	_unloadHandle	= unloadHandle;
	_server			= ioServer;
	_hdr			= hdr;
	_w = _h = 0;
	for( int i=0 ; i<_c ; i++ ) _w += _widths[i];
	for( int i=0 ; i<_r ; i++ ) _h += _heights[i];
}
template< class Real >
void MultiThreadedWGridOfImages< Real >::ImageWriter::Run( void )
{
	StreamingGrid** grids = new StreamingGrid*[_c];
	for( int i=0 ; i<_h ; i++ )
	{
		MyWaitForSingleObject( _unloadHandle , 10000 , "MultiThreadedWGridOfImages<Real>::ImageWriter::Run" );
		int row , x = 0 , y = i;
		for ( row = 0 ; y >= _heights[row] ; y -= _heights[row++] ) ;
		for( int col = 0 ; col < _c ; col++ )
		{

			if ( y==0 )
			{
				MyWinSock::StdinLock lock;
				grids[col] = GetWriteStream< Real >( _fileNames[col][row] , _widths[col] , _heights[row] , _separate , _quality , _server , _hdr );
			}
			if( !grids[col] )	fprintf(stderr,"Error: Attempting to write to NULL tile\n") , exit(0);
			Pointer( Real ) subRow = ( Pointer( Real ) )(*grids[col])[y];

			memcpy( subRow ,  _buffer+3*x , sizeof(Real)*_widths[col]*3 );

			x += _widths[col];
			grids[col]->advance();

			if( y >= grids[col]->rows()-1 ) delete grids[col] , grids[col]=NULL;
		}
		if( !SetEvent( _loadHandle ) ) fprintfId( stderr , "Set event failed: " ) , PrintError();
	}
	delete[] grids;
}

////////////////////////////////
// MultiThreadedWGridOfImages //
////////////////////////////////
template<class Real>
MultiThreadedWGridOfImages<Real>::MultiThreadedWGridOfImages( char* fileName , int width , int height , bool separateColors , int quality , int threads , int threadPriority , int tileWidth , int tileHeight , MultiStreamIOServer* ioServer , bool hdr )
{
	const char* ext = "jpg";
	char* header = GetFileHeader( fileName );
	_init( fileName , header , ext , width , height , separateColors , quality , threads , threadPriority , tileWidth , tileHeight , ioServer , hdr );
	delete[] header;
}
template<class Real>
MultiThreadedWGridOfImages<Real>::MultiThreadedWGridOfImages( char* fileName , const char* header , const char* ext , int width , int height , bool separateColors , int quality , int threads , int threadPriority , int tileWidth , int  tileHeight , MultiStreamIOServer* ioServer , bool hdr )
{
	_init( fileName , header , ext , width , height , separateColors , quality , threads , threadPriority , tileWidth , tileHeight , ioServer , hdr );
}
template<class Real>
void MultiThreadedWGridOfImages<Real>::_init( char* fileName , const char* header , const char* ext , int width , int height , bool separateColors , int quality , int threads , int threadPriority , int tileWidth , int tileHeight , MultiStreamIOServer* ioServer , bool hdr )
{
	FILE* fp = fopen(fileName,"w");
	_separate	= separateColors;
	_w			= width;
	_h			= height;
	_c			= (width + tileWidth -1) / tileWidth;
	_r			= (height + tileHeight -1) / tileHeight;
	_quality	= quality;
	_hdr		= hdr;
	_threads	= threads;
	_dataReady	= false;
	if( _threads>_c ) _threads = _c;
	char imageName[1024];
	fprintf( fp , "Columns: %d\n" , _c ) , fflush( fp );
	fprintf( fp , "Rows: %d\n" , _r )    , fflush( fp );

	int w , h;
	_widths  = new int[_c];
	_heights = new int[_r];
	for ( int i = 0 ; i < _c ; i++ )	_widths[i]  = tileWidth;
	for ( int j = 0 ; j < _r ; j++ )	_heights[j] = tileHeight;
	_widths[_c-1]  = width  - (_c-1)*tileWidth;
	_heights[_r-1] = height - (_r-1)*tileHeight;

	_fileNames = new char**[_c];
	for ( int i = 0 ; i < _c ; i++ ) _fileNames[i] = new char*[_r];
	for ( int j = 0 ; j < _r ; j++ )
		for ( int i = 0 ; i < _c ; i++ )
		{
			sprintf(imageName,"%s.%d.%d.%s",header,i,j,ext);
			_fileNames[i][j] = new char[strlen(imageName)+1];
			strcpy( _fileNames[i][j] , imageName );
			fprintf( fp , "%s\n" , imageName ) , fflush( fp );
		}
	fclose(fp);
	_buffer = AllocArray< Real >( _w * 3 , 1 , "MultiThreadedWGridOfImages::_buffer" );
	_imageWriterHandles	= new HANDLE[ _threads ];
	_loadHandles		= new HANDLE[ _threads ];
	_unloadHandles		= new HANDLE[ _threads ];
	_imageWriters		= new ImageWriter[ _threads ];
	int offset = 0;
	for( int i=0 ; i<_threads ; i++ )
	{
		_loadHandles[i]		= CreateEvent( NULL , false , true  , NULL );
		_unloadHandles[i]	= CreateEvent( NULL , false , false , NULL );
		int start = (i*_c) / _threads;
		int end = ( (i+1)*_c ) / _threads;
		_imageWriters[i].init( end-start , _r , _widths+start , _heights , _fileNames + start , _separate , quality , _buffer+offset*3 , _loadHandles[i] , _unloadHandles[i] , ioServer , hdr );
		for( int j=start ; j<end ; j++ ) offset += _widths[j];
		_imageWriterHandles[i] = SpawnThread< ImageWriter >( _imageWriters+i , threadPriority );
	}
	_current = 0;
}
template<class Real>
MultiThreadedWGridOfImages<Real>::~MultiThreadedWGridOfImages(void)
{
	delete[] _widths;
	delete[] _heights;
	_widths = _heights = NULL;
	for (int i = 0 ; i < _c ; i++ )
	{
		for ( int j = 0 ; j <_r ; j++ ) if( _fileNames[i][j] ) delete _fileNames[i][j];
		delete[] _fileNames[i];
	}
	delete[] _fileNames;
	FreeArray( _buffer );
	_fileNames = NULL;
	MyWaitForMultipleObjects( _threads , _imageWriterHandles , 10000 , "MultiThreadedWGridOfImages<Real>::~MultiThreadedWGridOfImages" );
	for( int i=0 ; i<_threads ; i++ )
	{
		CloseHandle( _loadHandles[i] );
		CloseHandle( _unloadHandles[i] );
	}
	delete[] _imageWriters;
	delete[] _imageWriterHandles;
	delete[] _loadHandles;
	delete[] _unloadHandles;
}
template<class Real> int MultiThreadedWGridOfImages<Real>::rows(void)	 const	{	return _h;	}
template<class Real> int MultiThreadedWGridOfImages<Real>::rowSize(void) const	{	return _w*3*sizeof(Real);	}
template<class Real> bool MultiThreadedWGridOfImages<Real>::SeparateColors(void)	const	{	return _separate;	}
template<class Real>
Pointer( byte ) MultiThreadedWGridOfImages<Real>::operator [] (int idx)
{
	if( !_dataReady )
	{
		MyWaitForMultipleObjects( _threads , _loadHandles , 10000 , "MultiThreadedWGridOfImages<Real>::operator[]" );
	}
	_dataReady = true;
	return ( Pointer( byte ) )( _buffer );
}
template<class Real>
void MultiThreadedWGridOfImages<Real>::advance(void)
{
	if( _current<_h )
	{
		if( !_dataReady )
		{
			MyWaitForMultipleObjects( _threads , _loadHandles , 10000 , "MultiThreadedWGridOfImages<Real>::advance" );
		}
		for( int i=0 ; i<_threads ; i++ ) if( !SetEvent( _unloadHandles[i] ) ) fprintfId( stderr , "SetEvent failed: " ) , PrintError();
		_dataReady = false;
	}
	_current++;
}

///////////////////
// WImageClipper //
///////////////////
template<class Real>
WImageClipper<Real>::WImageClipper( char* fileName , int width , int height , bool separateColors , int outOffX , int outOffY , int outWidth , int outHeight , int quality , MultiStreamIOServer* ioServer , bool hdr , int threads , int threadPriority )
{
	deleteGrid = true;
	_separate=separateColors;
	_hdr = hdr;
	_w = width;
	_h = height;
	_ox = outOffX;
	_oy = outOffY;
	_ow = outWidth;
	_oh = outHeight;

	if(_ox<0 || _oy<0 || _ox+_ow>_w || _oy+_oh>_h)
	{
		fprintf(stderr,"Clip-region not a subset of the domain in WImageClipper: [%d,%d) x [%d,%d) !< [0,%d) x [0,%d)\n",_ox,_ox+_ow,_oy,_oy+_oh,_w,_h);
		exit(0);
	}

	grid = GetWriteStream<Real>( fileName , _ow , _oh , false , quality , ioServer , hdr , threads , threadPriority );
	buffer = AllocArray< Real >( _w * 3 , 1 , "WImageClipper::buffer" );
	current = 0;
}
template<class Real>
WImageClipper<Real>::WImageClipper(StreamingGrid* outGrid,int width,int height,bool separateColors,int outOffX,int outOffY,int outWidth,int outHeight , bool hdr )
{
	deleteGrid = false;
	_separate=separateColors;
	_hdr = hdr;
	_w = width;
	_h = height;
	_ox = outOffX;
	_oy = outOffY;
	_ow = outWidth;
	_oh = outHeight;

	if(_ox<0 || _oy<0 || _ox+_ow>_w || _oy+_oh>_h)
	{
		fprintf(stderr,"Clip-region not a subset of the domain in WImageClipper: [%d,%d) x [%d,%d) !< [0,%d) x [0,%d)\n",_ox,_ox+_ow,_oy,_oy+_oh,_w,_h);
		exit(0);
	}

	grid = outGrid;
	buffer = AllocArray< Real >( _w * 3 , 1 , "WImageClipper::buffer" );
	current = 0;
}
template<class Real>
WImageClipper<Real>::~WImageClipper(void)
{
	if( deleteGrid ) delete grid;
	FreeArray( buffer );
}
template<class Real> int WImageClipper<Real>::rows(void)	const { return _h;}
template<class Real> int WImageClipper<Real>::rowSize(void)	const { return sizeof(Real)*_w*3;}
template<class Real> bool WImageClipper<Real>::SeparateColors(void)	const { return _separate; }

template<class Real>
Pointer( byte ) WImageClipper<Real>::operator [] (int idx)
{
	return ( Pointer( byte ) )( buffer );
}
template<class Real>
void WImageClipper<Real>::advance(void)
{
	if(current>=_oy && current<_oy+_oh)
	{
		Pointer( Real ) row = ( Pointer( Real ) )(*grid)[current-_oy];

		if(_separate)
		{
			for( int c=0 ; c<3 ;c++ )
				for( int i=0 ; i<_ow ; i++ )
					row[3*i+c] = buffer[_ox + i + c*_w];
		}
		else
			memcpy(row,buffer+3*_ox,sizeof(Real)*_ow*3);
		grid->advance();
	}
	current++;
}
///////////////////////
// RChannelExtractor //
///////////////////////
template<class Real>
RChannelExtractor<Real>::RChannelExtractor(char* fileName,int& width,int& height,bool separateColors,int channel)
{
	_c = channel;
	grid = GetReadStream<Real>(fileName,_w,_h,separateColors);
	width = _w;
	height = _h;

	buffer = new Real[_w];
	current = -1;
	advance();
}
template<class Real>
RChannelExtractor<Real>::~RChannelExtractor(void)
{
	delete[] buffer;
}
template<class Real>	int RChannelExtractor<Real>::rows(void)		const	{	return _h;	}
template<class Real>	int RChannelExtractor<Real>::rowSize(void)	const	{	return _w*sizeof(Real);	}
template<class Real>	bool RChannelExtractor<Real>::SeparateColors(void)	const	{	return _separate;	}
template<class Real>
Pointer( byte ) RChannelExtractor<Real>::operator [] (int idx)
{
	return ( Pointer( byte ) )( buffer , rowSize() );
}

template<class Real>
void RChannelExtractor<Real>::advance(void)
{
	current++;
	if(current<_h)
	{
		Real* row = (Real*)(*grid)[current];
		if(_separate)	memcpy(buffer,row+_c*_w,sizeof(Real)*_w);
		else	for(int x=0 ; x<_w ; x++)	buffer[x]=row[3*x+_c];
		grid->advance();
	}
}

///////////////////
// RImageSampler //
///////////////////
template< class Real >
RImageSampler< Real >::RImageSampler( char* fileName , int width , int  height , bool nearest , bool separateColors , MultiStreamIOServer* ioServer , int threads , int threadPriority )
{
	_grid = GetReadStream< Real >( fileName , _inW , _inH , separateColors , ioServer , threads , threadPriority);
	_outW = width;
	_outH = height;
	_separate = separateColors;
	_nearest = nearest;

	_outBuffer = AllocArray< Real >( 3*_outW , 1 );
	if( !nearest )
	{
		_inBuffers[0] = AllocArray< Real >( 3*_inW , 1 );
		_inBuffers[1] = AllocArray< Real >( 3*_inW , 1 );
	}
	else _inBuffers[0] = _inBuffers[1] = NullPointer< Real >( );
	_inCurrent = 0 , _outCurrent = -1;
	advance();
}
template< class Real >
RImageSampler<Real>::~RImageSampler( void )
{
	FreeArray( _outBuffer );
	FreeArray( _inBuffers[0] );
	FreeArray( _inBuffers[1] );
}
template< class Real > int  RImageSampler< Real >::rows( void ) const { return _outH; }
template< class Real > int  RImageSampler< Real >::rowSize(void) const { return _outW * sizeof( Real ) * 3; }
template< class Real > bool RImageSampler< Real >::SeparateColors( void ) const { return _separate; }
template< class Real >
Pointer( byte ) RImageSampler< Real >::operator [] ( int idx )
{
	return ( Pointer( byte ) ) _outBuffer;
}
template< class Real >
void RImageSampler< Real >::advance( void )
{
	_outCurrent++;
	float y = float( _outCurrent * ( _inH-1 ) ) / ( _outH-1 );
	int yy;
	if( _nearest )	// Align the input grid to the row we will be reading from
	{
		yy = int( y+0.5 );
		while( _inCurrent<yy && _inCurrent<_inH ) _grid->advance( ) , _inCurrent++;
	}
	else	// Align the input grid to the row we will be reading from and read in the row into the appropriate buffer
	{
		// This needs to be fixed!!!
		yy = int( y );
		while( _inCurrent<yy && _inCurrent<_inH ) _grid->advance( ) , _inCurrent++;
		if( _inCurrent==yy && _inCurrent<_inH )
		{
			Pointer( Real ) row = ( Pointer( Real ) )(*_grid)[_inCurrent];
			memcpy( _inBuffers[_inCurrent&1] , row , sizeof( Real ) * _inW * 3 );
			_grid->advance( );
			_inCurrent++;
		}
		if( _inCurrent==yy+1 && _inCurrent<_inH )
		{
			Pointer( Real ) row = ( Pointer( Real ) )(*_grid)[_inCurrent];
			memcpy( _inBuffers[_inCurrent&1] , row , sizeof( Real ) * _inW * 3 );
			_grid->advance( );
			_inCurrent++;
		}
	}
	if( _outCurrent<_outH )
	{
		if( _nearest )
		{
			Pointer( Real ) row = ( Pointer( Real ) )(*_grid)[_inCurrent];
			for( int i=0 ; i<_outW ; i++ )
			{
				float x = float( i * ( _inW-1 ) ) / ( _outW-1 );
				int xx = int( x+0.5 );
				if( _separate ) for( int c=0 ; c<3 ; c++ ) _outBuffer[i+_outW*c] = row[xx+_inW*c];
				else            for( int c=0 ; c<3 ; c++ ) _outBuffer[i*3+c] = row[3*xx+c];
			}
		}
		else
		{
			int y1 = yy;
			int y2 = y1+1;
			float dy = y-yy;
			for( int i=0 ; i<_outW ; i++ )
			{
				float x = float( i * ( _inW-1 ) ) / ( _outW-1 );
				int x1 = int( x );
				int x2 = x1 + 1;
				float dx = x-x1;
				if( x2>=_inW ) x2 = x1;
				if( _separate )
					for( int c=0 ; c<3 ; c++ )
						_outBuffer[i+_outW*c] =
						_inBuffers[y1&1][x1+_inW*c] * Real(1.0-dx) * Real(1.0-dy) +
						_inBuffers[y1&1][x2+_inW*c] * Real(    dx) * Real(1.0-dy) +
						_inBuffers[y2&1][x1+_inW*c] * Real(1.0-dx) * Real(    dy) +
						_inBuffers[y2&1][x2+_inW*c] * Real(    dx) * Real(    dy) ;
				else
					for( int c=0 ; c<3 ; c++ )
						_outBuffer[i*3+c] =
						_inBuffers[y1&1][x1*3+c] * Real(1.0-dx) * Real(1.0-dy) +
						_inBuffers[y1&1][x2*3+c] * Real(    dx) * Real(1.0-dy) +
						_inBuffers[y2&1][x1*3+c] * Real(1.0-dx) * Real(    dy) +
						_inBuffers[y2&1][x2*3+c] * Real(    dx) * Real(    dy) ;
			}
		}
	}
}

////////////////////////////////
#include <Util/cmdLineParser.h>
template<class Real>
void GetReadSize( char* fileName , int& width , int& height )
{
	char* ext = GetFileExtension( fileName );
	if( !strcasecmp( ext , "jpeg" ) || !strcasecmp( ext , "jpg" ) )	JPEGGetImageSize( fileName , width , height );
	else if( !strcasecmp( ext , "wdp" ) )							 WDPGetImageSize( fileName , width , height );
	else if( !strcasecmp( ext , "bmp" ) )							 BMPGetImageSize( fileName , width , height );
	else if( !strcasecmp( ext , "png" ) )							 PNGGetImageSize( fileName , width , height );
	else if( !strcasecmp( ext , "iGrid" ) )							RGridOfImages< Real >::GetImageSize( fileName , width , height );
	else if( !strcasecmp( ext , "int" ) )							GetImageSize< int     >( fileName , width , height );
	else if( !strcasecmp( ext , "int16" ) )							GetImageSize< __int16 >( fileName , width , height );
	else if( !strcasecmp( ext , "half" ) )							GetImageSize< half    >( fileName , width , height );
	else if( !strcasecmp( ext , "float" ) )							GetImageSize< float   >( fileName , width , height );
	else if( !strcasecmp( ext , "double" ) )						GetImageSize< double  >( fileName , width , height );
	else	fprintf( stderr , "Unsupported input file extension: %s\n" , ext );
	delete[] ext;
}

template<class Real>
StreamingGrid* GetReadStream( char* fileName , int& width , int& height , const bool& separateColors , MultiStreamIOServer* ioServer , int threads , int threadPriority  )
{
	StreamingGrid* data=NULL;
	char* ext=GetFileExtension(fileName);
	if(!strcasecmp(ext,"jpeg") || !strcasecmp(ext,"jpg") )	data = new JPEGRImageStream	<Real>			(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"wdp") )						data = new WDPRImageStream	<Real>			(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"bmp") )						data = new BMPRImageStream	<Real>			(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"png") )						data = new PNGRImageStream	<Real>			(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"pfm") )						data = new PFMRImageStream	<Real>			(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"iGrid") )
		if( !threads )										data = new RGridOfImages	<Real>			(fileName,width,height,separateColors , ioServer );
		else												data = new MultiThreadedRGridOfImages<Real>	( fileName , width , height , separateColors , threads , threadPriority , ioServer );
	else if(!strcasecmp(ext,"int") )						data = new RImageStream		<Real,int>		(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"int16") )						data = new RImageStream		<Real,__int16>	(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"half") )						data = new RImageStream		<Real,half>		(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"float") )						data = new RImageStream		<Real,float>	(fileName,width,height,separateColors , ioServer );
	else if(!strcasecmp(ext,"double") )						data = new RImageStream		<Real,double>	(fileName,width,height,separateColors , ioServer );
	else	fprintf(stderr,"Unsupported input file extension: %s\n",ext);
	delete[] ext;
	return data;
}
template<class Real>
StreamingGrid* GetWriteStream( char* fileName , const int& width , const int& height , const bool& separateColors , const int& quality , MultiStreamIOServer* ioServer , bool hdr , int threads , int threadPriority )
{
	StreamingGrid* data=NULL;
	char* ext=GetFileExtension(fileName);

	if(!strcasecmp(ext,"jpg") || !strcasecmp(ext,"jpeg"))	data=new JPEGWImageStream	<Real>			(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"wdp"))							data=new WDPWImageStream	<Real>			(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"bmp"))							data=new BMPWImageStream	<Real>			(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"png"))
		if( hdr )											data=new PNGWImageStreamHDR	<Real>			(fileName,width,height,separateColors,quality , ioServer );
		else												data=new PNGWImageStream	<Real>			(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"tiff"))
		if( hdr )											data=new TIFFWImageStreamHDR<Real>			(fileName,width,height,separateColors,quality , ioServer );
		else												data=new TIFFWImageStream	<Real>			(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"iGrid"))
		if( !threads )										data=new WGridOfImages		<Real>			( fileName , width , height , separateColors , quality , DefaultOutputTileWidth , DefaultOutputTileHeight , ioServer , hdr );
		else												data=new MultiThreadedWGridOfImages<Real>	( fileName , width , height , separateColors , quality , threads , threadPriority , DefaultOutputTileWidth , DefaultOutputTileHeight , ioServer , hdr );
	else if(!strcasecmp(ext,"int"))							data=new WImageStream		<Real,int>		(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"int16"))						data=new WImageStream		<Real,__int16>	(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"half"))						data=new WImageStream		<Real,half>		(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"float"))						data=new WImageStream		<Real,float>	(fileName,width,height,separateColors,quality , ioServer );
	else if(!strcasecmp(ext,"double"))						data=new WImageStream		<Real,double>	(fileName,width,height,separateColors,quality , ioServer );
	else	fprintf(stderr,"Unsupported output file extension: %s\n",ext);
	delete[] ext;
	return data;
}
