/*
Copyright (c) 2010, Michael Kazhdan
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the Johns Hopkins University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/
#include <stdlib.h>
#include <Util/TIFF/tiffio.h>
#include <Util/BaseMultiStreamIO.h>

struct TIFFWriteInfo
{
	TIFF* tiff;
};

void* TIFFInitWriteColor(char* fileName,int width,int height,int quality , MultiStreamIOServer* ioServer )
{
	TIFFWriteInfo* info = (TIFFWriteInfo*)malloc(sizeof(TIFFWriteInfo));

	info->tiff=TIFFOpen(fileName,"w");
	if (!info->tiff)															fprintf(stderr,"TIFFOpen failed\n")								,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGEWIDTH, width))					fprintf (stderr, "Can't set ImageWidth tag.\n")					,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGELENGTH, height))					fprintf (stderr, "Can't set ImageLength tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_BITSPERSAMPLE, sizeof(char)*8))		fprintf (stderr, "Can't set BitsPerSample tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_SAMPLESPERPIXEL, 3))					fprintf (stderr, "Can't set SamplesPerPixel tag.\n")			,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_COMPRESSION, COMPRESSION_LZW))		fprintf (stderr, "Can't set Compression tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ROWSPERSTRIP, 1))						fprintf (stderr, "Can't set RowsPerStrip tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG))	fprintf (stderr, "Can't set PlanarConfiguration tag.\n")		,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB))		fprintf (stderr, "Can't set PhotometricInterpretation tag.\n")	,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_RESOLUTIONUNIT, RESUNIT_NONE))		fprintf (stderr, "Can't set ResolutionUnit tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT))	fprintf (stderr, "Can't set Orientation tag.\n")				,	exit(0);

	return info;
}
void* TIFFInitWriteColorHDR(char* fileName,int width,int height,int quality , MultiStreamIOServer* ioServer )
{
	TIFFWriteInfo* info = (TIFFWriteInfo*)malloc(sizeof(TIFFWriteInfo));

	info->tiff=TIFFOpen(fileName,"w");
	if (!info->tiff)															fprintf(stderr,"TIFFOpen failed\n")								,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGEWIDTH, width))					fprintf (stderr, "Can't set ImageWidth tag.\n")					,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGELENGTH, height))					fprintf (stderr, "Can't set ImageLength tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_BITSPERSAMPLE, sizeof(__int16)*8))	fprintf (stderr, "Can't set BitsPerSample tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_SAMPLESPERPIXEL, 3))					fprintf (stderr, "Can't set SamplesPerPixel tag.\n")			,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_COMPRESSION, COMPRESSION_LZW))		fprintf (stderr, "Can't set Compression tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ROWSPERSTRIP, 1))						fprintf (stderr, "Can't set RowsPerStrip tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG))	fprintf (stderr, "Can't set PlanarConfiguration tag.\n")		,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB))		fprintf (stderr, "Can't set PhotometricInterpretation tag.\n")	,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_RESOLUTIONUNIT, RESUNIT_NONE))		fprintf (stderr, "Can't set ResolutionUnit tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT))	fprintf (stderr, "Can't set Orientation tag.\n")				,	exit(0);

	return info;
}
void* TIFFInitWriteGray(char* fileName,int width,int height,int quality , MultiStreamIOServer* ioServer )
{
	TIFFWriteInfo* info = (TIFFWriteInfo*)malloc(sizeof(TIFFWriteInfo));

	info->tiff=TIFFOpen(fileName,"w");
	if (!info->tiff)															fprintf(stderr,"TIFFOpen failed\n")								,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGEWIDTH, width))					fprintf (stderr, "Can't set ImageWidth tag.\n")					,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGELENGTH, height))					fprintf (stderr, "Can't set ImageLength tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_BITSPERSAMPLE, sizeof(char)*8))		fprintf (stderr, "Can't set BitsPerSample tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_SAMPLESPERPIXEL, 1))					fprintf (stderr, "Can't set SamplesPerPixel tag.\n")			,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_COMPRESSION, COMPRESSION_LZW))		fprintf (stderr, "Can't set Compression tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ROWSPERSTRIP, 1))						fprintf (stderr, "Can't set RowsPerStrip tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG))	fprintf (stderr, "Can't set PlanarConfiguration tag.\n")		,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK))	fprintf (stderr, "Can't set PhotometricInterpretation tag.\n")	,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_RESOLUTIONUNIT, RESUNIT_NONE))		fprintf (stderr, "Can't set ResolutionUnit tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT))	fprintf (stderr, "Can't set Orientation tag.\n")				,	exit(0);

	return info;
}
void* TIFFInitWriteGrayHDR(char* fileName,int width,int height,int quality , MultiStreamIOServer* ioServer )
{
	TIFFWriteInfo* info = (TIFFWriteInfo*)malloc(sizeof(TIFFWriteInfo));

	info->tiff=TIFFOpen(fileName,"w");
	if (!info->tiff)															fprintf(stderr,"TIFFOpen failed\n")								,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGEWIDTH, width))					fprintf (stderr, "Can't set ImageWidth tag.\n")					,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_IMAGELENGTH, height))					fprintf (stderr, "Can't set ImageLength tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_BITSPERSAMPLE, sizeof(__int16)*8))	fprintf (stderr, "Can't set BitsPerSample tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_SAMPLESPERPIXEL, 1))					fprintf (stderr, "Can't set SamplesPerPixel tag.\n")			,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_COMPRESSION, COMPRESSION_LZW))		fprintf (stderr, "Can't set Compression tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ROWSPERSTRIP, 1))						fprintf (stderr, "Can't set RowsPerStrip tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG))	fprintf (stderr, "Can't set PlanarConfiguration tag.\n")		,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK))	fprintf (stderr, "Can't set PhotometricInterpretation tag.\n")	,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_RESOLUTIONUNIT, RESUNIT_NONE))		fprintf (stderr, "Can't set ResolutionUnit tag.\n")				,	exit(0);
	if (!TIFFSetField(info->tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT))	fprintf (stderr, "Can't set Orientation tag.\n")				,	exit(0);

	return info;
}
void TIFFFinalizeWrite(void* v)
{
	TIFFWriteInfo* info = (TIFFWriteInfo*)v;
	TIFFClose(info->tiff);
	free(info);
}
void TIFFWriteRow(void* pixels,void* v,int j)
{
	unsigned __int16* _pixels = (unsigned __int16*)pixels;
	TIFFWriteInfo* info = (TIFFWriteInfo*)v;
	TIFFWriteScanline(info->tiff,pixels,j,0);
}
void TIFFWriteRowHDR(void* pixels,void* v,int j)
{
	unsigned char* _pixels = (unsigned char*)pixels;
	TIFFWriteInfo* info = (TIFFWriteInfo*)v;
	TIFFWriteScanline(info->tiff,pixels,j,0);
}
